# done
def fad_entire_calc_score(path, query_path, weights, search_features, a_s_f, query_features, a_q_f, clan_dict, option):
    """Main Scoring function, starts the functions for the scores and calculates the complete FAS score

    :param path: Path to score
    :param query_path: Path to score (query)
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param search_features: all features to be linearized in the seed protein
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param option: dictionary that contains the main option variables of FAS
    :return: score_ms, score_ps, score_cs, final_score, path_weight, common_feature, score_ls
    """
    score_cs = round(fad_entire_cs_score(path, query_path, query_features, clan_dict, search_features), 4)
    tmp = fad_entire_ms_score(path, query_path, search_features, a_s_f, query_features, a_q_f, weights, option)
    score_ms = round(tmp[0], 4)
    scale = tmp[2]
    path_weight = tmp[3]
    common_feature = tmp[4]
    ps_tmp = fad_entire_ps_score(path, tmp[2], query_path, search_features, a_s_f, query_features, a_q_f, weights,
                                 option)
    score_ps = round(ps_tmp[0], 4)
    score_ls = round(ps_tmp[1], 4)
    final_score = ((score_ms * option["score_weights"][0]) + (score_cs * option["score_weights"][1])
                   + (score_ps * option["score_weights"][2]))
    return score_ms, score_ps, score_cs, float(final_score), path_weight, common_feature, score_ls, scale


# done
def fad_entire_cs_score(path, query_path, query_features, clan_dict, search_features):
    """calculates clan score

    :param query_path:
    :param path: Path to score
    :param query_features: all features to be linearized in the query protein
    :param search_features: all features to be linearized in the seed protein
    :param clan_dict: dictionary that maps features to clans
    :return: score (PS)
    """

    counter = 0.0
    s_clans = {}  # clans from path
    q_clans = {}  # clans from query_path
    p_score = 0.0

    for i in path:
        if i in search_features:
            feature = search_features[i]
            if feature[0] in clan_dict:
                if clan_dict[feature[0]] in s_clans:
                    s_clans[clan_dict[feature[0]]] += 1
                else:
                    s_clans[clan_dict[feature[0]]] = 1
    for i in query_path:
        if i in query_features:
            feature = query_features[i]
            if feature[0] in clan_dict:
                if clan_dict[feature[0]] in q_clans:
                    q_clans[clan_dict[feature[0]]] += 1
                else:
                    q_clans[clan_dict[feature[0]]] = 1

    for clan in s_clans:
        counter += 1.0
        if clan in q_clans:
            p_score += 1.0 - min(float(s_clans[clan] * q_clans[clan]) / float(s_clans[clan] * s_clans[clan]), 1.0)
    if counter == 0:
        score = 0.0
    else:
        score = 1.0 - p_score / counter
    return float(score)


# done
def fad_entire_ms_score(path, query_path, search_features, a_s_f, query_features, a_q_f, weights, option):
    """Calculates multiplicity score

    :param path: Path to score
    :param query_path: Path to score (query)
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param search_features: all features to be linearized in the seed protein
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param option: dictionary that contains the main option variables of FAS
    :return: final_score (MS), search_domains, scale, final_weight, common_feature
    """
    search_domains = {}
    query_domains = {}
    scale = 0
    scores = []
    final_score = 1.0
    final_weight = 0.0
    main_features_s = {}
    main_features_q = []
    common_feature = False

    for i in path:
        if i in search_features:
            feature = search_features[i]
            main_features_s[feature[0]] = True
        else:
            feature = a_s_f[i]

        if feature[0] in search_domains:
            search_domains[feature[0]] += 1
        else:
            search_domains[feature[0]] = 1

    for i in query_path:
        if i in query_features:
            feature = query_features[i]
            main_features_q.append(feature[0])
        else:
            feature = a_q_f[i]
        if feature[0] in query_domains:
            query_domains[feature[0]] += 1
        else:
            query_domains[feature[0]] = 1

    # check for common features
    for y in main_features_q:
        try:
            if main_features_s[y]:
                common_feature = True
        except KeyError:
            pass
    for feature in search_domains:
        if feature in weights['weights']:
            scale += weights['weights'][feature]
        else:
            scale += weights['default']
        if feature in query_domains:
            p_score = 1.0 - min(float(search_domains[feature] * query_domains[feature]) /
                                float(search_domains[feature] * search_domains[feature]), 1.0)
            scores.append((feature, p_score))
        else:
            scores.append((feature, 1.0))
    if scale > 1.0:
        scale = 1.0 / float(scale)
    else:
        scale = 1.0
    for score in scores:
        if score[0] in weights['weights']:
            weight = weights['weights'][score[0]]
        else:
            weight = weights['default']
        final_score -= score[1] * scale * weight
        final_weight += weight
    return float(final_score), search_domains, scale, final_weight, common_feature


#done
def fad_entire_ps_score(path, scale, query_path, search_features, a_s_f, query_features, a_q_f, weights, option):
    """Calculates positional score FAD variant

    :param path: Path to score
    :param scale: contains scaling factor for each weight or number of feature types for uniform weighting
    :param query_path: Path to score (query)
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param search_features: all features to be linearized in the seed protein
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param option: dictionary that contains the main option variables of FAS
    :return: final_score (PS), final_ls_score
    """
    count = {}
    final_score = 1.0
    final_ls_score = 1.0
    scores = {}
    ls_scores = {}
    local_query_protein = {}
    ls = 0.0
    # get current features from query path
    for i in query_path:
        if i in query_features:
            feature = query_features[i]
        else:
            feature = a_q_f[i]
        if feature[0] in local_query_protein:
            length = feature[3] - feature[2] + 1.0
            local_query_protein[feature[0]].append((feature[1], length))
        else:
            local_query_protein[feature[0]] = [(feature[1], feature[3] - feature[2] + 1.0)]

    # compare features in path with features form query path
    for i in path:
        if i in search_features:
            feature = search_features[i]
        else:
            feature = a_s_f[i]
        if feature[0] in local_query_protein:
            min_distance = 1.0
            if not feature[0] in scores:
                count[feature[0]] = 0
                scores[feature[0]] = 0.0
                ls_scores[feature[0]] = 0.0
            for instance in local_query_protein[feature[0]]:
                distance = float(abs(feature[1] - instance[0]))
                if min_distance > distance:
                    min_distance = distance
                    if instance[1] <= (feature[3] - feature[2]):
                        ls = 1 - float(instance[1]) / float(feature[3] - feature[2] + 1.0)
                    else:
                        ls = 1 - float(feature[3] - feature[2] + 1.0) / float(instance[1])

            scores[feature[0]] += min_distance
            ls_scores[feature[0]] += ls
            count[feature[0]] += 1
        else:
            scores[feature[0]] = 1.0
            ls_scores[feature[0]] = 1.0
            count[feature[0]] = 1
    for f_score in scores:
        if f_score in weights['weights']:
            weight = weights['weights'][f_score]
        else:
            weight = weights['default']
        final_score -= scores[f_score] / count[f_score] * scale * weight
        final_ls_score -= ls_scores[f_score] / count[f_score] * scale * weight
    return float(final_score), float(final_ls_score)



#!/bin/env python

#######################################################################
# Copyright (C) 2022 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import argparse
from pkg_resources import get_distribution
from greedyFAS.mainFAS.fasInput import read_json


def get_anno_version(path):
    anno = read_json(path)
    if 'version' in anno:
        for tool in anno['version']:
            print(tool)
            for x in anno['version'][tool]:
                print(x + ': ' + str(anno['version'][tool][x]))
            print('-----------------------------------')
    else:
        print('This seems to be a pre 1.15 annotation file. Specific version data is only for annotations with FAS version '
              '1.15 and higher.')


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    required.add_argument("-i", "--input", default=None, type=str, required=True,
                          help="path to annotation json")
    args = parser.parse_args()
    get_anno_version(args.input)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of FAS.
#  Update annotation file with InterPro Acc, phmm length and tool versions
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import sys
import os
import argparse
from pathlib import Path
import greedyFAS.annoFAS.annoModules as annoModules
from pkg_resources import get_distribution


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-a', '--annoFile', help='Input annotation file in json format', action='store', default='',
                          required=True)
    optional.add_argument('-o', '--outPath', help='Output directory', action='store', default='')
    optional.add_argument('--silent', help='Turn off terminal output', action='store_true')
    optional.add_argument('--annoToolFile', help='Path to files contains annotation tool names',
                          action='store', default='')

    args = parser.parse_args()

    annoFile = args.annoFile
    annoModules.checkFileExist(annoFile)
    annoFile = os.path.abspath(annoFile)
    outPath = os.path.abspath(args.outPath)
    Path(outPath).mkdir(parents=True, exist_ok=True)
    silent = args.silent
    annoToolFile = args.annoToolFile
    annoModules.checkFileExist(annoToolFile)

    try:
        annoModules.updateAnnoFile(annoFile)
        annoModules.printMsg(silent, '%s updated successfully!' % annoFile)
    except:
        print('ERROR: Cannot update %s! You can check with fas.checkAnno or contact us!' % annoFile)

if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2023 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import argparse
import json
from plotly.subplots import make_subplots
import plotly.graph_objects as go
from pkg_resources import get_distribution


def get_stats(annotation_path, tool1, tool2, threshold, svg):
    with open(annotation_path, 'r') as infile:
        in_dict = json.loads(infile.read())
    statistics = count_overlaps(in_dict['feature'], tool1, tool2, threshold)
    fig = make_subplots(rows=3, cols=2,
                        column_titles=[tool1, tool2],
                        row_titles=['Domains', 'Domain Instances (overlap)', 'Domain Instances (all)'],
                        specs=[[{'type': 'domain'}, {'type': 'domain'}],
                               [{'type': 'domain'}, {'type': 'domain'}],
                               [{'type': 'domain'}, {'type': 'domain'}]])
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[0], statistics[1]-statistics[0]]),
        row=1, col=1
    )
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[4], statistics[5]-statistics[4]]),
        row=1, col=2
    )
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[8], statistics[9]-statistics[8]]),
        row=2, col=1
    )
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[10], statistics[11]-statistics[10]]),
        row=2, col=2
    )
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[2], statistics[3]-statistics[2]]),
        row=3, col=1
    )
    fig.add_trace(
        go.Pie(labels=['Overlapping', 'Overlap Free'],
               values=[statistics[6], statistics[7]-statistics[6]]),
        row=3, col=2
    )
    fig.show()
    if svg:
        out = '.'.join(annotation_path.split('.')[0:-1])
        fig.write_image(file=out + "_Overlap_Statistics.svg", width=800, height=800)


def count_overlaps(proteome, t1, t2, threshold):
    t1_f_o = {}  # dictionary of features types with number of overlaps t1
    t1_f_a = {}  # set of all features in t2
    t1_o_total = 0  # total number of feature instances that have at least one overlap in t1
    t1_f_total = 0  # total number of feature instance in t1
    t1_rel_total = 0  # number of feature instances that have at least one overlap out of overlapping types
    t1_rel_a_total = 0  # number of feature instances out of overlapping types
    t2_f_o = {}
    t2_f_a = {}
    t2_f_total = 0
    t2_o_total = 0
    t2_rel_total = 0
    t2_rel_a_total = 0
    for protein in proteome:
        t1_f_a, t1_f_total, t1_f_o, t1_o_total = count_overlaps_sub(proteome[protein], t1_f_a, t1_f_total, t1_f_o,
                                                                    t1_o_total, t1, t2, threshold)
        t2_f_a, t2_f_total, t2_f_o, t2_o_total = count_overlaps_sub(proteome[protein], t2_f_a, t2_f_total, t2_f_o,
                                                                    t2_o_total, t2, t1, threshold)
    for feature in t1_f_o:
        t1_rel_total += t1_f_o[feature]
        t1_rel_a_total += t1_f_a[feature]
    for feature in t2_f_o:
        t2_rel_total += t2_f_o[feature]
        t2_rel_a_total += t2_f_a[feature]
    return (len(t1_f_o), len(t1_f_a), t1_o_total, t1_f_total, len(t2_f_o), len(t2_f_a), t2_o_total, t2_f_total,
            t1_rel_total, t1_rel_a_total, t2_rel_total, t2_rel_a_total)


def count_overlaps_sub(arc, f_a, f_total, f_o, o_total, t1, t2, threshold):
    for feature in arc[t1]:
        if feature not in f_a:
            f_a[feature] = 0
        for instance in arc[t1][feature]['instance']:
            f_a[feature] += 1
            f_total += 1
            start, stop = instance[0], instance[1]
            length = stop - start
            if count_overlaps_sub2(arc[t2], start, stop, length, threshold):
                if feature not in f_o:
                    f_o[feature] = 0
                f_o[feature] += 1
                o_total += 1
    return f_a, f_total, f_o, o_total


def count_overlaps_sub2(arc, start, stop, length, threshold):
    back = False
    for feature in arc:
        for instance in arc[feature]['instance']:
            length2 = instance[1] - instance[0]
            if instance[0] <= start <= instance[1]:
                o_length = instance[1] - start
                if (length2 * threshold) <= o_length >= (length * threshold):
                    back = True
            elif start <= instance[0] <= stop:
                o_length = stop - instance[0]
                if (length2 * threshold) <= o_length >= (length * threshold):
                    back = True
    return back


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("-i", "--input", default=None, type=str, required=True,
                          help="path to input annotation file [.json]")
    optional.add_argument("-a", "--class1", default='pfam', type=str, required=False,
                          help="Evaluated feature class 1")
    optional.add_argument("-b", "--class2", default='smart', type=str, required=False,
                          help="Evaluated feature class 2")
    optional.add_argument("-t", "--threshold", default=0.8, type=float, required=False,
                          help="Threshold for overlap size in comparison to feature lengths. The overlap must at least "
                               "reach the given fraction [0.0 to 1.0] of both feature lengths to be considered.")
    optional.add_argument("-s", "--svg", action="store_true",
                          help="Save as .svg. Will be saved in same directory as input.")
    args = parser.parse_args()
    get_stats(args.input, args.class1, args.class2, args.threshold, args.svg)


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2019 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


from os.path import abspath


def write_tsv_out(outpath, bidirectional, results):
    out = open(outpath + '.tsv', 'w')
    outdict = {}
    out.write('Seed\tQuery\tScore(Forward/Reverse)\tMS(Forward/Reverse)\tPS(Forward/Reverse)\tCS(Forward/Reverse)'
              '\tLS(Forward/Reverse)\tMethod\n')
    for result in results[0]:
        outdict[result[0], result[1]] = (result[2], ('NA', 'NA', 'NA', 'NA', 'NA'), result[3])
    if bidirectional:
        for result in results[1]:
            outdict[result[1], result[0]] = (outdict[result[1], result[0]][0], result[2], outdict[result[1],
                                                                                                  result[0]][2])
    for pair in outdict:
        out.write(pair[0] + '\t' + pair[1] + '\t' + f'{outdict[pair][0][0]:.4}' + '/'
                  + f'{outdict[pair][1][0]:.4}' + '\t' + f'{outdict[pair][0][1]:.4}' + '/'
                  + f'{outdict[pair][1][1]:.4}' + '\t' + f'{outdict[pair][0][2]:.4}' + '/'
                  + f'{outdict[pair][1][2]:.4}' + '\t' + f'{outdict[pair][0][3]:.4}' + '/'
                  + f'{outdict[pair][1][3]:.4}' + '\t' + f'{outdict[pair][0][4]:.4}' + '/'
                  + f'{outdict[pair][1][4]:.4}' + '\t' + outdict[pair][2] + '\n')
    out.close()


def write_domain_out_fad(seed_proteome, query_proteome, seed, query, weights, scale, seedpath, querypath, out, option,
                         interprokeys, phmm):
    tools = option['input_linearized'] + option['input_normal']
    if option['reverse']:
        groupname = query
    else:
        groupname = seed
    for tool in tools:
        for feature in seed_proteome[seed][tool]:
            interpro = 'NA'
            if feature in interprokeys:
                interpro = interprokeys[feature].split('.')[0]
            if feature in seedpath:
                if feature in weights['weights']:
                    weight = round(weights['weights'][feature] * scale, 4)
                else:
                    weight = round(weights['default'] * scale, 4)
                for instance in seed_proteome[seed][tool][feature]['instance']:
                    if (instance[0], instance[1]) in seedpath[feature]:
                        inpath = 'Y'
                    else:
                        inpath = 'N'
                    if len(instance) > 3:
                        phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) + '\t' \
                                    + str(instance[5]) + '\t'
                    else:
                        phmm_info= '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                    if feature in phmm:
                        phmm_info = phmm_info + str(phmm[feature]) + '\n'
                    else:
                        phmm_info = phmm_info + 'NA\n'

                    if option['reverse']:
                        out.write(groupname + '#' + seed + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\t'
                                  + str(weight) + '\t' + inpath + '\t' + interpro + phmm_info)
                    else:
                        out.write(groupname + '#' + query + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\t'
                                  + str(weight) + '\t' + inpath + '\t' + interpro + phmm_info)
            else:
                for instance in seed_proteome[seed][tool][feature]['instance']:
                    if len(instance) > 3:
                        phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) + '\t' \
                                    + str(instance[5]) + '\t'
                    else:
                        phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                    if feature in phmm:
                        phmm_info = phmm_info + str(phmm[feature]) + '\n'
                    else:
                        phmm_info = phmm_info + 'NA\n'
                    if option['reverse']:
                        out.write(groupname + '#' + seed + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\tNA\tN\t'
                                  + interpro + phmm_info)
                    else:
                        out.write(groupname + '#' + query + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\tNA\tN\t'
                                  + interpro + phmm_info)
    if not seed == query:
        for tool in tools:
            for feature in query_proteome[query][tool]:
                interpro = 'NA'
                if feature in interprokeys:
                    interpro = interprokeys[feature].split('.')[0]
                if feature in seedpath:
                    if feature in weights['weights']:
                        weight = round(weights['weights'][feature] * scale, 4)
                    else:
                        weight = round(weights['default'] * scale, 4)
                else:
                    weight = 'NA'
                if feature in querypath:
                    for instance in query_proteome[query][tool][feature]['instance']:
                        if (instance[0], instance[1]) in querypath[feature]:
                            inpath = 'Y'
                        else:
                            inpath = 'N'
                        if len(instance) > 3:
                            phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) \
                                        + '\t' + str(instance[5]) + '\t'
                        else:
                            phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                        if feature in phmm:
                            phmm_info = phmm_info + str(phmm[feature]) + '\n'
                        else:
                            phmm_info = phmm_info + 'NA\n'
                        if option['reverse']:
                            out.write(groupname + '#' + seed + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\t' + inpath + '\t' + interpro
                                      + phmm_info)
                        else:
                            out.write(groupname + '#' + query + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\t' + inpath + '\t' + interpro
                                      + phmm_info)
                else:
                    for instance in query_proteome[query][tool][feature]['instance']:
                        if len(instance) > 3:
                            phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4])\
                                        + '\t' + str(instance[5]) + '\t'
                        else:
                            phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                        if feature in phmm:
                            phmm_info = phmm_info + str(phmm[feature]) + '\n'
                        else:
                            phmm_info = phmm_info + 'NA\n'
                        if option['reverse']:
                            out.write(groupname + '#' + seed + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\tN\t' + interpro
                                      + phmm_info)
                        else:
                            out.write(groupname + '#' + query + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\tN\t' + interpro
                                      + phmm_info)


def write_domain_out(seed_proteome, query_proteome, seed, query, weights, scale, seedpath, querypath, out, option,
                     interprokeys, phmm):
    tools = option['input_linearized'] + option['input_normal']
    uni_weight = None
    if option['reverse']:
        groupname = query
    else:
        groupname = seed
    if option['MS_uni'] and len(seedpath) > 0:
        uni_weight = round(1.0 / len(seedpath), 4)
    for tool in tools:
        for feature in seed_proteome[seed][tool]:
            interpro = 'NA'
            if feature in interprokeys:
                interpro = interprokeys[feature].split('.')[0]
            if feature in seedpath:
                if option['MS_uni']:
                    weight = uni_weight
                else:
                    weight = round(weights[feature] * scale, 4)
                for instance in seed_proteome[seed][tool][feature]['instance']:
                    if (instance[0], instance[1]) in seedpath[feature]:
                        inpath = 'Y'
                    else:
                        inpath = 'N'
                    if len(instance) > 3:
                        phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) + '\t' \
                                    + str(instance[5]) + '\t'
                    else:
                        phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                    if feature in phmm:
                        phmm_info = phmm_info + str(phmm[feature]) + '\n'
                    else:
                        phmm_info = phmm_info + 'NA\n'
                    if option['reverse']:
                        out.write(groupname + '#' + seed + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\t'
                                  + str(weight) + '\t' + inpath + '\t' + interpro + phmm_info)
                    else:
                        out.write(groupname + '#' + query + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\t'
                                  + str(weight) + '\t' + inpath + '\t' + interpro + phmm_info)
            else:
                for instance in seed_proteome[seed][tool][feature]['instance']:
                    if len(instance) > 3:
                        phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) + '\t' \
                                    + str(instance[5]) + '\t'
                    else:
                        phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                    if feature in phmm:
                        phmm_info = phmm_info + str(phmm[feature]) + '\n'
                    else:
                        phmm_info = phmm_info + 'NA\n'
                    if option['reverse']:
                        out.write(groupname + '#' + seed + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\tNA\tN\t'
                                  + interpro + phmm_info)
                    else:
                        out.write(groupname + '#' + query + '\t' + seed + '\t' + str(seed_proteome[seed]['length'])
                                  + '\t' + feature + '\t' + str(instance[0]) + '\t' + str(instance[1]) + '\tNA\tN\t'
                                  + interpro + phmm_info)
    if not seed == query:
        for tool in tools:
            for feature in query_proteome[query][tool]:
                interpro = 'NA'
                if feature in interprokeys:
                    interpro = interprokeys[feature].split('.')[0]
                if feature in seedpath:
                    if option['MS_uni']:
                        weight = uni_weight
                    else:
                        weight = round(weights[feature] * scale, 4)
                else:
                    weight = 'NA'
                if feature in querypath:
                    for instance in query_proteome[query][tool][feature]['instance']:
                        if (instance[0], instance[1]) in querypath[feature]:
                            inpath = 'Y'
                        else:
                            inpath = 'N'
                        if len(instance) > 3:
                            phmm_info = ('\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4])
                                         + '\t' + str(instance[5]) + '\t')
                        else:
                            phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                        if feature in phmm:
                            phmm_info = phmm_info + str(phmm[feature]) + '\n'
                        else:
                            phmm_info = phmm_info + 'NA\n'
                        if option['reverse']:
                            out.write(groupname + '#' + seed + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\t' + inpath + '\t' + interpro
                                      + phmm_info)
                        else:
                            out.write(groupname + '#' + query + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\t' + inpath + '\t' + interpro
                                      + phmm_info)
                else:
                    for instance in query_proteome[query][tool][feature]['instance']:
                        if len(instance) > 3:
                            phmm_info = '\t' + str(instance[2]) + '\t' + str(instance[3]) + '\t' + str(instance[4]) + '\t'\
                                        + str(instance[5]) + '\t'
                        else:
                            phmm_info = '\t' + str(instance[2]) + '\tNA\tNA\tNA\t'
                        if feature in phmm:
                            phmm_info = phmm_info + str(phmm[feature]) + '\n'
                        else:
                            phmm_info = phmm_info + 'NA\n'
                        if option['reverse']:
                            out.write(groupname + '#' + seed + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\tN\t' + interpro
                                      + phmm_info)
                        else:
                            out.write(groupname + '#' + query + '\t' + query + '\t' +
                                      str(query_proteome[query]['length']) + '\t' + feature + '\t' + str(instance[0])
                                      + '\t' + str(instance[1]) + '\t' + str(weight) + '\tN\t' + interpro
                                      + phmm_info)


def phyloprofile_out(outpath, bidirectional, mapping_file, results):
    with open(mapping_file) as infile:
        map = {}
        for line in infile.readlines():
            cells = line.rstrip('\n').split('\t')
            map[cells[0]] = cells[1]
    out = open(outpath + '.phyloprofile', 'w')
    out.write('geneID\tncbiID\torthoID\tFAS_F\tFAS_B\n')
    outdict = {}

    for result in results[0]:
        outdict[result[0], result[1]] = (result[2][0], 0.0)
    if bidirectional:
        for result in results[1]:
            outdict[result[1], result[0]] = (outdict[result[1], result[0]][0], result[2][0])
    for pair in outdict:
        try:
            out.write(pair[0] + '\t' + map[pair[1]] + '\t' + pair[1] + '\t' + f'{outdict[pair][0]:.4}' + '\t'
                      + f'{outdict[pair][1]:.4}' + '\n')
        except KeyError:
            raise Exception(pair[1] + ' not in mapping file')
    out.close()


def write_metadata(path, args, commandline, version):
    paths = ['seed', 'query', 'annotation_dir', 'out_dir', 'toolPath', 'ref_proteome', 'ref_2', 'weight_constraints',
             'pairwise', 'featuretypes', 'phyloprofile']
    with open(path, 'w') as out:
        out.write('# Metadata file generated automatically by FAS\nFAS_version: '
                  + version + '\n' + 'command_line: ' + commandline + '\n')
        arguments = vars(args)
        for argument in arguments:
            if argument in paths and arguments[argument]:
                out.write(argument + ': ' + abspath(arguments[argument]) + '\n')
            else:
                out.write(argument + ': ' + str(arguments[argument]) + '\n')



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2021 Vinh Tran
#
#  This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import sys
import os
import argparse
from pathlib import Path
import subprocess
from Bio import SeqIO
import shutil
import multiprocessing as mp
import json
import greedyFAS.annoFAS.annoModules as annoModules
from pkg_resources import get_distribution


def checkCompleteAnno(seqFile, jsonFile):
    allSeq = []
    allSeqDict = SeqIO.to_dict((SeqIO.parse(open(seqFile), 'fasta')))
    allSeq = list(allSeqDict.keys())
    with open(jsonFile) as jf:
        dt = json.load(jf)
        annotated = dt['feature']
        missingSeqs = list(set(allSeq).difference(annotated))
        if len(missingSeqs) > 0:
            out = []
            for seq in missingSeqs:
                out.append('>%s\n%s\n' % (seq, allSeqDict[seq].seq))
            return(out)
        else:
            return()


def checkOutdatedAnno(jsonFile):
    with open(jsonFile) as jf:
        annoDict = json.load(jf)
        if not 'interproID' in annoDict:
            return('interproID')
        elif not 'version' in annoDict:
            return('version')
        elif 'version' in annoDict:
            tools = list(annoDict['version'].keys())
            toolPath = annoModules.getFasToolPath()
            cutoffs = (0.001, 0.01, 0.0000001, 'euk')
            currentVer = annoModules.getVersions(tools, toolPath, cutoffs)
            outdatedTools = []
            for tool in tools:
                if 'version' in currentVer[tool]:
                    if 'version' in annoDict['version'][tool]:
                        if not annoDict['version'][tool]['version'] == currentVer[tool]['version']:
                            outdatedTools.append('%s (%s vs %s)' % (tool, annoDict['version'][tool]['version'], currentVer[tool]['version']))
                    else:
                        return('version')
            if len(outdatedTools) > 0:
                return(', '.join(outdatedTools))
            else:
                return('updated')
        else:
            return('updated')


def doAnnoForMissing(taxon, missingAnno, jsonFile, outPath, cpus, silent, annoToolFile):
    toolPath = annoModules.getFasToolPath()
    # do annotation for missing proteins
    faFile = '%s/%s_tmp.fa' % (outPath, taxon)
    with open(faFile, 'w') as f:
        f.write(''.join(missingAnno))

    annoCmd = 'fas.doAnno -i %s -o %s --cpus %s' % (faFile, outPath, cpus)
    if annoToolFile:
        annoCmd += ' -annoToolFile %s' % annoToolFile
    if silent:
        annoCmd = annoCmd + ' > /dev/null 2>&1'
    try:
        subprocess.call([annoCmd], shell = True)
    except:
        print('\033[91mProblem with running fas.doAnno. You can check it with this command:\n%s\033[0m' % annoCmd)
    # merge with old annotation json file
    with open(jsonFile) as oldJson:
        oldAnno = json.load(oldJson)
    with open('%s/%s_tmp.json' % (outPath, taxon)) as newJson:
        newAnno = json.load(newJson)
    annoOut = [oldAnno, newAnno]
    annoDict = annoModules.mergeNestedDic(annoOut)
    annoDict['clan'] = annoModules.getClans(toolPath, annoDict['feature'])
    annoDict['count'] = annoModules.countFeatures(annoDict['feature'])
    annoModules.save2json(annoDict, taxon, outPath)
    # remove tmp files
    os.remove(faFile)
    os.remove('%s/%s_tmp.json' % (outPath, taxon))
    shutil.rmtree('%s/tmp' % outPath)


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-s', '--seqFile', help='Input sequence file in fasta format', action='store', default='',
                          required=True)
    required.add_argument('-a', '--annoFile', help='Input annotation file in json format', action='store', default='',
                          required=True)
    required.add_argument('-o', '--outPath', help='Output directory', action='store', default='', required=True)
    optional.add_argument('-n', '--noAnno', help='Do not annotate missing proteins', action='store_true')
    optional.add_argument('-u', '--update', help='Update data structure of annotation file (not the annotations themselves)', action='store_true')
    optional.add_argument('--keep', help='Keep old annotaion file', action='store_true')
    optional.add_argument('--silent', help='Turn off terminal output', action='store_true')
    optional.add_argument('--cpus', help='Number of CPUs used for annotation. Default = available cores - 1',
                          action='store', default=0, type=int)
    optional.add_argument('--annoToolFile', help='Path to files contains annotation tool names',
                          action='store', default='')

    args = parser.parse_args()

    seqFile = args.seqFile
    annoModules.checkFileExist(seqFile)
    seqFile = os.path.abspath(seqFile)
    annoFile = args.annoFile
    annoModules.checkFileExist(annoFile)
    annoFile = os.path.abspath(annoFile)
    outPath = os.path.abspath(args.outPath)
    Path(outPath).mkdir(parents=True, exist_ok=True)
    cpus = args.cpus
    if cpus == 0:
        cpus = mp.cpu_count()-1
    noAnno = args.noAnno
    update = args.update
    keep = args.keep
    silent = args.silent
    annoToolFile = args.annoToolFile
    annoModules.checkFileExist(annoToolFile)

    taxon = annoFile.split('/')[-1].replace('.json', '')

    ### check for missing annotation
    missingAnno = checkCompleteAnno(seqFile, annoFile)
    if len(missingAnno) > 0:
        if noAnno == False:
            doAnnoForMissing(taxon, missingAnno, annoFile, outPath, cpus, silent, annoToolFile)
            annoModules.printMsg(silent, '%s missing proteins of %s has been annotated!' % (len(missingAnno), taxon))
        else:
            print('WARNING: Annotation for %s proteins of %s are missing!\n%s' % (len(missingAnno), taxon, missingAnno))
    else:
        annoModules.printMsg(silent, 'Annotations found for all %s sequences!' % (taxon))


    ### check for outdated annotation file
    outdatedCheck = checkOutdatedAnno(annoFile)
    if outdatedCheck == 'updated':
        annoModules.printMsg(silent, 'Annotation file updated!')
    elif outdatedCheck == 'interproID' or outdatedCheck == 'version':
        if update == True:
            oldAnnoFile = annoModules.updateAnnoFile(annoFile)
            annoModules.printMsg(silent, '%s updated!' % annoFile)
            if keep == False:
                os.remove(oldAnnoFile)
        else:
            print('WARNING: %s is not updated! Please consider update it with --update option' % annoFile)
    else:
        print('WARNING: Annotations from the tools below are outdated. Please consider to re-annotate them!\n%s' % outdatedCheck)



if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2022 Julian Dosch
#
# This file is part of greedyFAS.
#
#  fas_disorder_addon is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  greedyFAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with greedyFAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


from Bio import SeqIO
import subprocess
import os
import sys
from pathlib import Path
import multiprocessing as mp
from tqdm import tqdm
import argparse
import json
from pkg_resources import get_distribution


def save2json(dict2save, outpath):
    jsonOut = json.dumps(dict2save, ensure_ascii=False)
    f = open(outpath, 'w')
    f.write(jsonOut)
    f.close()


def parse_aucpred(infile):
    struc = []
    with open(infile, 'r') as toparse:
        for line in toparse.readlines():
            if not line[0] == '#':
                if line.split()[2] == '*':
                    struc.append(1)
                elif line.split()[2] == '.':
                    struc.append(0)
                else:
                    raise Exception('Disordered annotation seems to be faulty!')
    dis = []
    switch = False
    start = None
    for i in range(len(struc)):
        if struc[i] and not switch:
            start = i
            switch = True
        elif not struc[i] and switch:
            dis.append([start, i, 'NA'])
            switch = False
    return {'aucpred': {'aucpred_disordered_region': {'evalue': 'NA', 'instance': dis}}}


def prepare_annojobs(infile, tmppath, aucpred):
    annojobs = []
    ids = []
    i = 0
    for seq in SeqIO.parse(infile, 'fasta'):
        if seq.id in ids:
            raise Exception('Headers in the fasta are not unique.')
        else:
            ids.append(seq.id)
            annojobs.append([seq.id, str(seq.seq), tmppath, aucpred, 'seq_' + str(i)])
        i += 1
    return annojobs


def remove_tmp(filepath):
    if os.path.exists(filepath):
        os.remove(filepath)


def make_tmp_fasta(header, seq, path):
    with open(path + '/' + header + '.fasta', 'w') as out:
        out.write('>sequence\n' + seq)


def run_anno(inpath, outpath, tmppath, cpus, aucpred):
    name = ''.join(inpath.split('/')[-1].split('.')[0:-1])
    Path(outpath).mkdir(parents=True, exist_ok=True)
    Path(tmppath + name + '/').mkdir(parents=True, exist_ok=True)
    joblist = prepare_annojobs(inpath, tmppath + name + '/', aucpred)
    out = []
    pool = mp.Pool(cpus)
    try:
        for _ in tqdm(pool.imap_unordered(run_anno_single, joblist), total=len(joblist), mininterval=5.0):
            out.append(_)
    except subprocess.CalledProcessError:
        raise 'Something went wrong during annotation.'
    pool.close()
    write_output(out, outpath + '/' + name)
    Path(tmppath + name + '/').rmdir()


def write_output(out, outpath):
    failed = []
    count = 0
    fdict = {}
    for seq in out:
        if not seq[1]:
            failed.append(seq[0])
        else:
            fdict[seq[0]] = seq[1]
            count += len(seq[1]['aucpred']['aucpred_disordered_region']['instance'])
    if not len(failed) == len(out):
        outdict = {'feature': fdict, 'clan': {}, 'count': {'aucpred_disordered_region': count}}
        save2json(outdict, outpath + '_disorder.json')
    if failed:
        print("Errors occured for the annotation of the following sequences:")
        for i in failed:
            print(i)


def run_anno_single(args):
    header, seq, tmppath, aucpred, tmpname = args
    make_tmp_fasta(tmpname, seq, tmppath)
    try:
        disorder = run_aucpred(tmpname, tmppath, aucpred)
        if os.path.exists(tmppath + tmpname + '.fasta'):
            os.remove(tmppath + tmpname + '.fasta')
        return header, disorder
    except:
        if os.path.exists(tmppath + tmpname + '.fasta'):
            os.remove(tmppath + tmpname + '.fasta')
        return header, None


def run_aucpred(header, tmppath, aucpred):
    cmd = aucpred + ' -i "' + tmppath + header + '.fasta" -o ' + tmppath
    subprocess.run([cmd], shell=True, capture_output=True, check=True)
    disorder = parse_aucpred(tmppath + header + '.diso_noprof')
    for i in ['.diso_noprof', '.diso_prev']:
        remove_tmp(tmppath + header + i)
    return disorder


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument("-i", "--input", default='.', type=str, required=True,
                          help="path to input fasta")
    required.add_argument("-o", "--outPath", default='.', type=str, required=True,
                          help="path to output directory. File will be named after fasta file")
    optional.add_argument("-t", "--tmp", default='.', type=str, required=False,
                          help="Path to a temporary directory.")
    optional.add_argument("--cpus", default=1, type=int, required=False,
                          help="number of cpus used")
    optional.add_argument("-a", "--aucpred", default=None, type=str, required=False,
                          help="Alternative path to AUCpreD.sh.")
    args = parser.parse_args()
    if args.aucpred:
        toolpath = args.aucpred
    else:
        pathconfigFile = os.path.realpath(__file__).replace('disorderFAS/predict_disorder.py', 'pathconfig.txt')
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.txt found. Please run fas.setup (https://github.com/BIONF/FAS/wiki/setup).')
        with open(pathconfigFile, 'r') as input:
            toolpath = input.readline().strip() + '/Predict_Property/AUCpreD.sh'
    if not os.path.exists(toolpath):
        sys.exit('AUCpreD.sh not found either give a path to the AUCpreD.sh with [-a|--aucpred] or install via '
                 '"fas.disorder.setup".')
    run_anno(args.input, args.outPath, args.tmp, args.cpus, toolpath)


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2020 Vinh Tran
#
#  This file is part of FAS.
#  This script is used for preparing the annotation tools for annoFAS
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import os
import sys
from sys import platform
from pathlib import Path
import shutil
import subprocess
import argparse
import gnureadline
import glob
from os.path import expanduser
import ssl
import urllib.request
import time
import multiprocessing as mp
from tqdm import tqdm
from greedyFAS.disorderFAS import install_aucpred
import greedyFAS.annoFAS.annoModules as annoModules
from pkg_resources import get_distribution

home = expanduser('~')


def complete(text, state):
    return(glob.glob(os.path.expanduser(text)+'*')+[None])[state]


def subprocess_cmd(commands):
    for cmd in commands:
        subprocess.call(cmd, shell=True)


def download_progress(count, block_size, total_size):
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = int(progress_size / (1024 * duration))
    percent = int(count * block_size * 100 / total_size)
    if percent > 100:
        percent = 100
    sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed" %
                     (percent, progress_size / (1024 * 1024), speed, duration))
    sys.stdout.flush()


def download_file(url, file):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    download_file = urllib.request.URLopener(context=ctx)
    print('Downloading %s' % (url + '/' + file))
    download_file.retrieve(url + '/' + file, file, download_progress)
    print(' ... done!')


def download_data(file, checksum, toolPath):
    url = 'https://applbio.biologie.uni-frankfurt.de/download/hamstr_qfo'
    download_file(url, file)
    if os.path.isfile(file):
        checksum_file = subprocess.check_output(['cksum', file]).decode(sys.stdout.encoding).strip()
        if checksum_file == checksum:
            print('Extracting %s ...' % file)
            shutil.unpack_archive(file, toolPath, 'gztar')
        else:
            sys.exit('ERROR: Downloaded file corrupted!')
    else:
        sys.exit('Cannot download annotation tools!')


def query_yes_no(question, default='yes'):
    valid = {'yes': True, 'y': True, 'ye': True,
             'no': False, 'n': False}
    if default is None:
        prompt = ' [y/n] '
    elif default == 'yes':
        prompt = ' [Y/n] '
    elif default == 'no':
        prompt = ' [y/N] '
    else:
        raise ValueError('invalid default answer: "%s"' % default)
    while True:
        choice = sys.stdin.readline().rstrip().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write('Please respond with "yes" or "no" '
                             '(or "y" or "n").\n')


def check_conda_env():
    """ Return if a conda env is currently using """
    if 'CONDA_DEFAULT_ENV' in os.environ:
        if not os.environ['CONDA_DEFAULT_ENV'] == 'base':
            return(True)
    return(False)


def get_dtu_path(dtuPathIn):
    if not dtuPathIn == '':
        if not os.path.isdir(dtuPathIn):
            sys.exit('ERROR: ' + dtuPathIn + ' not found!')
        else:
            dtu_path = os.path.abspath(dtuPathIn)
    else:
        print('Do you want to use TMHMM (v2.0c) and SignalP (v4.1g)? (y/n)')
        if query_yes_no('Source path?'):
            gnureadline.set_completer_delims(' \t\n;')
            gnureadline.parse_and_bind('tab: complete')
            gnureadline.set_completer(complete)
            dtu_path = input('Please download TMHMM 2.0c and SignalP 4.1g at https://services.healthtech.dtu.dk/ and '
                             'enter path to the downloaded tar files:')
        else:
            dtu_path = ''

    if not dtu_path == '':
        files = [f for f in os.listdir(dtu_path) if os.path.isfile(os.path.join(dtu_path, f))]
        signalp_source = [i for i in files if 'signalp' in i]
        tmhmm_source = [i for i in files if 'tmhmm' in i]
        if not tmhmm_source:
            sys.exit('ERROR: TMHMM not found in %s' % dtu_path)
        if not signalp_source:
            sys.exit('ERROR: signalp-4.1g not found in %s' % dtu_path)
        return dtu_path, signalp_source[0], tmhmm_source[0]
    else:
        return 0, 0, 0


def check_status(toolPath, force, reinstall):
    flag = 1
    cwd = os.getcwd()
    if os.path.isfile(toolPath+'/annoTools.txt'):
        with open(toolPath+'/annoTools.txt') as f:
            if '#checked' in f.read():
                if force:
                    print('Annotation tools found in %s will be deleted and reinstalled! Enter to continue.' % toolPath)
                    if query_yes_no(''):
                        try:
                            shutil.rmtree(toolPath)
                        except:
                            sys.exit('ERROR: Failed to delete %s. Please manually remove it and run fas.setup again!'
                                     % toolPath)
                        flag = 1
                    else:
                        flag = toolPath
                else:
                    if len(reinstall) > 0:
                        flag = 1
                    else:
                        flag = toolPath
        f.close()
    return flag


def install_signalp():
    shutil.move('signalp-4.1', 'SignalP')
    subprocess.call(['chmod', '-R', '+w', './SignalP'])
    os.chdir('SignalP')
    signalp_path = os.getcwd()
    with open(signalp_path+'/signalp', "r") as infile, open(signalp_path+'/signalp.mod', "w") as outfile:
        for line in infile:
            if 'ENV{SIGNALP} = ' in line:
                line = '    $ENV{SIGNALP} = "%s";\n' % signalp_path
            if 'uname -m' in line:
                if 'Debian' in os.uname()[3]:
                    line = 'my $architecture = \"i686\";\n'
            outfile.write(line)
    shutil.move(signalp_path+'/signalp.mod', signalp_path+'/signalp')
    subprocess.call(['chmod', '0755', signalp_path+'/signalp'])


def install_tmhmm():
    shutil.move('tmhmm-2.0c', 'TMHMM')
    machine = os.uname()[4]
    if 'Debian' in os.uname()[3]:
        machine = 'i686'
    os.chdir('TMHMM')
    makelink_tmhmm = 'ln -s -f bin/decodeanhmm.Linux_%s decodeanhmm' % machine
    subprocess.call([makelink_tmhmm], shell=True)


def correct_aln(args):
    (aln_file, anno_path) = args
    tmp = {}
    file_name = aln_file.split('/')[-1].split('.')[0]
    Path(anno_path + '/SMART/aln').mkdir(parents = True, exist_ok = True)
    with open(aln_file, 'r') as f:
        for l in [line.rstrip() for line in f]:
            if not l.startswith('CLUSTAL') and not l.startswith(' ') and len(l) > 0:
                if not l.startswith('>'):
                    id = l.split()[0].replace('.', '_')
                    if len(l.split()) > 1:
                        if not id in tmp:
                            tmp[id] = l.split()[1].strip()
                        else:
                            tmp[id] += l.split()[1].strip()
                else:
                    break
    if not len(tmp) > 0:
        if not os.path.exists(f'{anno_path}/SMART/aln/{file_name}.aln'):
            os.symlink(aln_file, f'{anno_path}/SMART/aln/{file_name}.aln')
    else:
        with open(f'{anno_path}/SMART/aln/{file_name}.aln', 'w') as o:
            for i in tmp:
                o.write(f'>{i}\n{tmp[i]}\n')


def create_phmm(args):
    (aln_file, anno_path) = args
    file_name = aln_file.split('/')[-1].split('.')[0]
    Path(anno_path + '/SMART/hmm').mkdir(parents = True, exist_ok = True)
    if os.path.exists(f'{anno_path}/SMART/aln/{file_name}.aln'):
        try:
            hmmbuildCmd = f'hmmbuild {anno_path}/SMART/hmm/{file_name}.hmm {anno_path}/SMART/aln/{file_name}.aln > /dev/null 2>&1'
            subprocess.run([hmmbuildCmd], shell=True, check=True)
        except:
            print(f'ERROR: Problem occurred while creating HMM profile for {aln_file}!')


def install_smart(smart_path, anno_path):
    aln_files = glob.glob(f'{smart_path}/aln/*.aln')
    if len(aln_files) < 1:
        sys.exit(f'ERROR: Cannot find aln files in {smart_path}')
    if not os.path.exists('%s/SMART/SMART-hmms/SMART.hmm.length' % anno_path):
        # create profile hmms from aln files
        aln_jobs = []
        hmm_jobs = []
        for aln in aln_files:
            aln_jobs.append([aln, anno_path])
            hmm_jobs.append([aln, anno_path])
        cpus = mp.cpu_count()-1
        pool = mp.Pool(cpus)
        for _ in tqdm(pool.imap_unordered(correct_aln, aln_jobs), total = len(aln_jobs)):
            pass
        for _ in tqdm(pool.imap_unordered(create_phmm, hmm_jobs), total = len(hmm_jobs)):
            pass
        pool.close()
        # create hmm database for hmmscan
        Path(anno_path + '/SMART/SMART-hmms').mkdir(parents=True, exist_ok=True)
        catCmd = f'cat {anno_path}/SMART/hmm/*.hmm > {anno_path}/SMART/SMART-hmms/SMART.hmm'
        subprocess.call([catCmd], shell=True)
        hmmpressCmd = 'hmmpress -f %s/SMART/SMART-hmms/SMART.hmm' % (anno_path)
        try:
            subprocess.run([hmmpressCmd], shell=True, check=True)
        except:
            sys.exit('ERROR: Problem occurred while creating binary files for SMART at %s/SMART/SMART-hmms' % (anno_path))
        getVersionCmd = 'tail -n 1 %s/README | cut -d " " -f2 | sed "s/\//_/g" > %s/SMART/version.txt' % (smart_path, anno_path)
        subprocess.call([getVersionCmd], shell=True)
        # write length file
        write_phmm_length(anno_path, 'SMART')
        # remove temp folders (aln and hmm)
        shutil.rmtree(f'{anno_path}/SMART/aln')
        shutil.rmtree(f'{anno_path}/SMART/hmm')
    else:
        print('WARNING: SMART is already installed!')


def install_pfam(pfam_version, pfam_path, anno_path):
    if not os.path.exists('%s/Pfam/Pfam-hmms/Pfam-A.hmm.length' % anno_path):
        file = 'Pfam-A.hmm.gz'
        dat_file = 'Pfam-A.hmm.dat.gz'
        relnotes_file = 'relnotes.txt'
        if not pfam_path:
            url = 'https://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/'
            if pfam_version:
                url = 'https://ftp.ebi.ac.uk/pub/databases/Pfam/releases/Pfam%s' % pfam_version
            download_file(url, file)
            download_file(url, dat_file)
            download_file(url, relnotes_file)
        else:
            annoModules.checkFileExist(f'{pfam_path}/{file}')
            annoModules.checkFileExist(f'{pfam_path}/{dat_file}')
            annoModules.checkFileExist(f'{pfam_path}/{relnotes_file}')
        if os.path.isfile(file):
            Path(anno_path + '/Pfam/Pfam-hmms').mkdir(parents=True, exist_ok=True)
            # move donwloaded file to anno_path/Pfam/Pfam-hmms
            try:
                if not pfam_path:
                    shutil.move(file, '%s/Pfam/Pfam-hmms/%s' % (anno_path, file))
                    shutil.move(dat_file, '%s/Pfam/Pfam-hmms/%s' % (anno_path, dat_file))
                    shutil.move(relnotes_file, '%s/Pfam/Pfam-hmms/%s' % (anno_path, relnotes_file))
                else:
                    shutil.copy(f'{pfam_path}/{file}', f'{anno_path}/Pfam/Pfam-hmms/{file}')
                    shutil.copy(f'{pfam_path}/{dat_file}', f'{anno_path}/Pfam/Pfam-hmms/{dat_file}')
                    shutil.copy(f'{pfam_path}/{relnotes_file}', f'{anno_path}/Pfam/Pfam-hmms/{relnotes_file}')
            except:
                sys.exit('ERROR: Cannot move/copy %s, %s & %s to %s/Pfam/Pfam-hmms' % (file, dat_file, relnotes_file, anno_path))
            # unzip downloaded files
            unzipCmd = 'gzip -d %s/Pfam/Pfam-hmms/*.gz' % anno_path
            try:
                subprocess.run([unzipCmd], shell=True, check=True)
            except:
                sys.exit('ERROR: Cannot upzip gz files in %s/Pfam/Pfam-hmms' % anno_path)
            # create bin files for Pfam-A.hmm
            hmmpressCmd = 'hmmpress -f %s/Pfam/Pfam-hmms/Pfam-A.hmm' % (anno_path)
            try:
                subprocess.run([hmmpressCmd], shell=True, check=True)
            except:
                sys.exit('ERROR: Problem occurred while creating binary files for PFAM at %s/Pfam/Pfam-hmms' % (anno_path))
            # write length file for PFAM
            write_phmm_length(anno_path, 'Pfam')
        else:
            sys.exit('ERROR: No Pfam-A.hmm.gz found at %s' % url)
    else:
        print('WARNING: PFAM is already installed!')


def write_phmm_length(anno_path, tool_name):
    hmm_file_name = f'{tool_name}-A.hmm'
    if tool_name == 'SMART':
        hmm_file_name = f'{tool_name}.hmm'
    hmm_file = f'{anno_path}/{tool_name}/{tool_name}-hmms/{hmm_file_name}'
    with open(f'{anno_path}/{tool_name}/{tool_name}-hmms/{hmm_file_name}.length', 'w') as len_f:
        with open(hmm_file, 'r') as file:
            for l in [line.rstrip() for line in file]:
                if l.startswith('NAME'):
                    id = l.split()[1].strip()
                if l.startswith('LENG'):
                    len_f.write(f'{id}\t{l.split()[1].strip()}\n')


def install_coils(anno_path):
    if not os.path.islink('%s/COILS2/COILS2' % anno_path):
        # url = 'http://ftp.ebi.ac.uk/pub/software/unix/coils-2.2/'
        # file = 'ncoils.tar.gz'
        url = 'http://www.russelllab.org/coils/'
        file = 'coils.tar.gz'
        download_file(url, file)
        if os.path.isfile(file):
            coils_path = anno_path + '/COILS2'
            Path(coils_path).mkdir(parents=True, exist_ok=True)
            # move donwloaded file to anno_path/COILS2
            try:
                shutil.move(file, '%s/%s' % (coils_path, file))
            except:
                sys.exit('ERROR: Cannot move %s to %s' % (file, coils_path))
            # re-compile COILS2
            os.chdir(coils_path)
            shutil.unpack_archive(file, coils_path, 'gztar')
            coils_bin = coils_path + '/coils'
            os.chdir(coils_bin)
            compile_cmd = 'cc -O2 -I. -o ncoils-osf ncoils.c read_matrix.c -lm' + ' > /dev/null 2>&1'
            try:
                subprocess.call(compile_cmd, shell=True)
            except:
                os.chdir(anno_path)
                sys.exit('ERROR: Failed to compile COILS2.\nPlease read instruction at %s and do it manually!' % coils_path)
            os.chdir(coils_path)
            if os.path.exists('COILS2'):
                os.remove('COILS2')
            os.symlink('coils/ncoils-osf', './COILS2')
            os.symlink('coils/new.mat', './new.mat')
            os.symlink('coils/old.mat', './old.mat')
            os.chdir(anno_path)
            COILSDIR = 'export COILSDIR=%s\n' % coils_bin
            with open(anno_path + '/fas.profile', 'w') as outfile:
                outfile.write(COILSDIR)
    else:
        print('WARNING: COILS2 is already installed!')


def install_flps(anno_path):
    if not os.path.islink('%s/fLPS/fLPS' % anno_path):
        # download flps
        fLPS_file = 'fLPS2programs.tar' #'fLPS.tar.gz'
        fLPS_url = 'http://biology.mcgill.ca/faculty/harrison/'
        download_file(fLPS_url, fLPS_file)
        shutil.unpack_archive(fLPS_file, anno_path, 'gztar')
        os.remove(fLPS_file)
        os.chdir(anno_path)
        shutil.move('fLPS2programs', 'fLPS')
        # compile and make symlink for fLPS
        flps_path = anno_path + '/fLPS'
        shutil.rmtree(anno_path + '/fLPS/bin')
        os.chdir(flps_path + '/src')
        makeCmd = 'make'
        os.mkdir(anno_path + '/fLPS/bin')
        try:
            subprocess.call([makeCmd], shell=True)
            for i in ['CompositionMaker', 'DomainFilter', 'fLPS2']:
                shutil.move(flps_path + '/src/' + i, flps_path + '/bin/' + i)
            source = flps_path + '/bin/fLPS2'
            target = flps_path + '/fLPS'
            try:
                os.symlink(source, target)
            except FileExistsError:
                os.remove(target)
                os.symlink(source, target)
        except:
            os.chdir(anno_path)
            sys.exit('ERROR: Failed to compile fLPS.\nPlease try to do it manually at' % flps_path)
        os.chdir(anno_path)
    else:
        print('WARNING: fLPS is already installed!')


def install_seg(anno_path):
    if not os.path.islink('%s/SEG/seg' % anno_path):
        url = 'https://raw.githubusercontent.com/BIONF/data4travis/main'
        file = 'SEG.tar.gz'
        download_file(url, file)
        shutil.unpack_archive(file, anno_path, 'gztar')
        os.remove(file)
    else:
        print('WARNING: SEG is already installed!')


def check_hmmer():
    try:
        subprocess.check_output(['which hmmsearch'], shell = True, stderr = subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if check_conda_env() == True:
            conda_install_cmd = 'conda install -c bioconda hmmer -y'
            try:
                subprocess.call([conda_install_cmd], shell = True)
            except:
                sys.exit('\033[91mERROR: Cannot install hmmer using this command\n%s\033[0m' % conda_install_cmd)
        else:
            sys.exit('\033[91mERROR: Please install HMMER before using FAS (http://hmmer.org/documentation.html)\033[0m')


def check_installed_tools(anno_path, ignoreList, reinstall, force, greedyFasPath):
        # list of annotation tools
        defaultTools = ['Pfam', 'fLPS', 'COILS2', 'SEG']
        if platform == 'darwin':
            ignoreList.extend(('SEG', 'TMHMM'))
        tools = [tool for tool in defaultTools if tool not in ignoreList]

        # check if anno tools exist
        if os.path.exists(os.path.abspath(anno_path+'/annoTools.txt')):
            with open(anno_path+'/annoTools.txt') as checkfile:
                if '#checked' in checkfile.read():
                    if not force:
                        if len(reinstall) > 0:
                            # get existing tools
                            tools = []
                            with open(anno_path+'/annoTools.txt') as f:
                                file =  f.readlines()
                                for line in file:
                                    if ('#' not in line) and (len(line) > 1):
                                        tools.append(line.strip())
                            # delete installed tools to reinstall
                            for r_tool in reinstall:
                                if os.path.exists('%s/%s' % (anno_path, r_tool)):
                                    print('Deleting %s/%s' % (anno_path, r_tool))
                                    shutil.rmtree('%s/%s' % (anno_path, r_tool))
                                    if not r_tool in tools:
                                        tools.append(r_tool)
                        else:
                            print('Annotation tools already found at %s' % anno_path)
                            if not os.path.exists(os.path.abspath(greedyFasPath + '/pathconfig.txt')):
                                print('If you want to add %s to config file of FAS, please rerun this function with --savePath!'
                                      % anno_path)
                            print('If you want to re-install them, rerun this function with --force!')
                            sys.exit()
        return(tools)


def install_annoTool(options):
    anno_path = options['toolPath']
    dtuPathIn = options['dtuPath']
    smart_path = options['smartPath']
    force = options['force']
    ignoreList = options['ignore']
    reinstall = options['reinstall']
    pfam_version = options['pfamVersion']
    pfam_path = options['pfamPath']
    check_hmmer()


    tools = check_installed_tools(anno_path, ignoreList, reinstall, force, options['greedyFasPath'])

    # do install anno tools
    current_dir = os.getcwd()
    if check_status(anno_path, force, reinstall) == 1:
        Path(anno_path).mkdir(parents=True, exist_ok=True)
        anno_path = os.path.abspath(anno_path)
        os.chdir(anno_path)

        print('Annotation tools will be installed in ')
        print(os.getcwd())
        print('----------------------------------')
        # install TMHMM and SignalP
        if len(reinstall) == 0 or 'SignalP' in reinstall or 'TMHMM' in reinstall:
            if not ('TMHMM' in ignoreList and 'SignalP' in ignoreList):
                (dtu_path, signalp_source, tmhmm_source) = get_dtu_path(dtuPathIn)
            else:
                dtu_path = 0
            if not dtu_path == 0:
                if not 'SignalP' in ignoreList:
                    print('==> Installing SignalP...')
                    if not os.path.isdir(anno_path + '/SignalP'):
                        shutil.unpack_archive(dtu_path + '/' + signalp_source, anno_path, 'gztar')
                        install_signalp()
                    os.chdir(anno_path)
                    if not 'SignalP' in tools:
                        tools.append('SignalP')
                if not 'TMHMM' in ignoreList:
                    if not platform == 'darwin':
                        print('==> Installing TMHMM...')
                        if not os.path.isdir(anno_path + '/TMHMM'):
                            shutil.unpack_archive(dtu_path + '/' + tmhmm_source, anno_path, 'gztar')
                            install_tmhmm()
                        os.chdir(anno_path)
                        if not 'TMHMM' in tools:
                            tools.append('TMHMM')

        # install SMART
        if len(reinstall) == 0 or 'SMART' in reinstall:
            if not smart_path == '':
                print('==> Installing SMART...')
                install_smart(smart_path, anno_path)
                if not 'SMART' in tools:
                    tools.append('SMART')
            else:
                if 'SMART' in reinstall:
                    sys.exit('ERROR: You must provide your downloaded SMART data. Check https://github.com/BIONF/FAS/wiki/setup for more details!')

        # install PFAM
        if len(reinstall) == 0 or 'Pfam' in reinstall:
            if pfam_path:
                print('==> Installing PFAM from %s...' % pfam_path)
            elif pfam_version:
                print('==> Installing PFAM version %s...' % pfam_version)
            else:
                print('==> Installing PFAM latest version...')
            install_pfam(pfam_version, pfam_path, anno_path)

        # install fLPS, COILS2 and SEG
        if len(reinstall) == 0 or 'fLPS' in reinstall:
            if not 'fLPS' in ignoreList:
                print('==> Installing fLPS ...')
                install_flps(anno_path)

        if len(reinstall) == 0 or 'COILS2' in reinstall:
            if 'COILS2' not in ignoreList:
                print('==> Installing COILS2...')
                install_coils(anno_path)

        if len(reinstall) == 0 or 'SEG' in reinstall:
            if 'SEG' not in ignoreList:
                print('==> Installing SEG...')
                install_seg(anno_path)

        # install disorder
        if options['disorder']:
            install_aucpred.install_auc(options['toolPath'])

        # write annoTools file
        with open('annoTools.txt', mode='wt') as tool_file:
            tool_file.write('#linearized\nPfam\n')
            if 'SMART' in tools:
                tool_file.write('SMART\n')
            tool_file.write('#normal\n')
            for t in tools:
                if not t in ['Pfam','SMART']:
                    tool_file.write('%s\n' % t)
        tool_file.close()

        print('----------------------------------')
        os.chdir(current_dir)
        return anno_path
    else:
        os.chdir(current_dir)
        return check_status(anno_path, force, reinstall)


def checkExecutable(anno_path):
    with open(anno_path+'/annoTools.txt') as file:
        availTool = [line.strip() for line in file]
    sedCmd ='perl -pi -e \'s/#checked\\n//\' %s/annoTools.txt' % anno_path
    subprocess.call([sedCmd], shell=True)

    print('Checking if annotation tools are excutable...')
    # test pfam and smart
    if not os.path.isfile(anno_path + '/Pfam/Pfam-hmms/Pfam-A.hmm'):
        sys.exit('Pfam hmm file not found. Please run fas.setup with --force!')
    if 'SMART' in availTool:
        if not os.path.isfile(anno_path + '/SMART/SMART-hmms/SMART.hmm'):
            sys.exit('SMART hmm file not found. Please run fas.setup with --force!')
    # test seg
    if 'SEG' in availTool:
        segCmd = '%s/SEG/seg' % anno_path
        try:
            p1 = subprocess.Popen([segCmd], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output1, err1 = p1.communicate()
            if not 'Usage' in err1.decode('UTF-8').strip():
                sys.exit('Error with SEG. You can reinstall it by running fas.setup with --force!')
        except:
            sys.exit('Error with SEG. You can reinstall it by running fas.setup with --force!')
    # test fLPS
    if 'fLPS' in availTool:
        flpsCmd = '%s/fLPS/fLPS' % anno_path
        try:
            p2 = subprocess.Popen([flpsCmd], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output2, err2 = p2.communicate()
            if not err2.decode('UTF-8').strip() == 'There is no sequence file. Please supply one in FASTA format.': # 'There is no sequence file. Please supply one.':
                sys.exit('Error with fLPS. You can reinstall it by running fas.setup with --force!')
        except:
            sys.exit('Error with fLPS. You can reinstall it by running fas.setup with --force!')
    # test COILS2
    if 'COILS2' in availTool:
        profileFile = '%s/fas.profile' % anno_path
        if not os.path.isfile(profileFile):
            sys.exit('No config file for COILSDIR found. Please check https://github.com/BIONF/FAS/wiki/FAQ#Error-with-COILS2!')
    # test tmhmm
    if 'TMHMM' in availTool:
        tmhmmCmd = '%s/TMHMM/decodeanhmm' % anno_path
        try:
            p4 = subprocess.Popen([tmhmmCmd], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output4, err4 = p4.communicate()
            if not 'Error: No modelfile given' in err4.decode('UTF-8').strip():
                sys.exit('Error with TMHMM. You can reinstall it by running fas.setup with --force!')
        except:
            sys.exit('Error with TMHMM. You can reinstall it by running fas.setup with --force!')
    # test signalp
    if 'SignalP' in availTool:
        signalpCmd = '%s/SignalP/signalp' % anno_path
        try:
            p5 = subprocess.Popen([signalpCmd, '-V'], stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
            output5, err5 = p5.communicate()
            if not err5.decode('UTF-8').strip() == '':
                sys.exit('Error with SignalP. Please check https://github.com/BIONF/FAS/wiki/FAQ#Error-with-SignalP!')
        except:
            sys.exit('Error with SignalP. Please check https://github.com/BIONF/FAS/wiki/FAQ#Error-with-SignalP!')
    return True


def checkAnnoToolsFile(toolPath):
    if not os.path.exists(os.path.abspath(toolPath+'/annoTools.txt')):
        sys.exit('ERROR: %s not found' % (toolPath+'/annoTools.txt'))
    else:
        with open(toolPath+'/annoTools.txt') as f:
            if not '#checked' in f.read():
                sys.exit('ERROR: Some errors occur with annotation tools. Please install them again!')


def saveConfigFile(checkResult, anno_path, greedyFasPath):
    if checkResult:
        with open(anno_path+'/annoTools.txt') as f:
            if '#checked' not in f.read():
                with open(anno_path+'/annoTools.txt', 'a') as file:
                    file.write('#checked')
                    file.close()
        f.close()
        with open(greedyFasPath+'/pathconfig.txt', 'w') as config:
            config.write(os.path.abspath(anno_path))
            config.close()
        print('Done! Annotation tools can be found in %s' % anno_path)
        print('You should test fas.doAnno with this command:')
        print('==> fas.doAnno -i test_annofas.fa -o testFas_output <==')
        print('*** NOTE: YOU NEED TO source %s/fas.profile BEFORE USING FAS!' % anno_path)
        print('Check https://github.com/BIONF/FAS/wiki/setup#add-COILSDIR-to-bashrc-or-bash_profile for more details')
        sys.exit()
    else:
        sys.exit('Some errors occur with annotation tools. Please check if they can be excuted at %s' % anno_path)


def main():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.')
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-t', '--toolPath', help='Set path to save annotation tools', action='store', default='',
                          required=True)
    optional.add_argument('--smartPath', help='Set path to your downloaded SMART folder', action='store', default='')
    optional.add_argument('-d', '--dtuPath', help='Set path to DTU tools (SignalP and TMHMM)', action='store',
                          default='')
    optional.add_argument('--pfamVersion', help='Specify PFAM version. E.g.: 35.0', action='store', default='')
    optional.add_argument('--pfamPath', help='Set path to your downloaded PFAM folder', action='store', default='')
    optional.add_argument('-i', '--ignore', help='List of tools should be ignored, e.g. SEG COILS2', nargs="*",
                        choices=['SignalP', 'TMHMM', 'COILS2', 'fLPS', 'SEG'],
                        action='store', default=[])
    optional.add_argument('-f', '--force', help='Overwrite old annotation tools if exist', action='store_true')
    optional.add_argument('-r', '--reinstall', help='List of tools should be reinstalled, e.g. SMART SignalP', nargs="*",
                        choices=['SMART','Pfam','SignalP', 'TMHMM', 'COILS2', 'fLPS', 'SEG'],
                        action='store', default=[])
    optional.add_argument('-s', '--savePath', help='Save annotation tool path to config file for FAS',
                          action='store_true')
    optional.add_argument('-c', '--check', help='Check if FAS ready to run. NO real tool path need to be given!',
                          action='store_true')
    optional.add_argument('--checkExecutable', help='Check if annotation tools are executable!', action='store_true')
    optional.add_argument('--addLength', help='Create length files for PFAM and SMART', action='store_true')
    optional.add_argument('--disorder', help='Install Disorder annotation, can be installed separately later as well',
                          action='store_true')

    args = parser.parse_args()
    greedyFasPath = os.path.realpath(__file__).replace('/setupFAS.py','')
    options = {
        'toolPath': args.toolPath,
        'smartPath': args.smartPath,
        'dtuPath': args.dtuPath,
        'pfamVersion': args.pfamVersion,
        'pfamPath': args.pfamPath,
        'ignore': args.ignore,
        'force': args.force,
        'reinstall': args.reinstall,
        'greedyFasPath': greedyFasPath,
        'disorder': args.disorder
    }

    if args.checkExecutable:
        if not os.path.exists(os.path.abspath(args.toolPath+'/annoTools.txt')):
            sys.exit('ERROR: %s not found' % (args.toolPath+'/annoTools.txt'))
        else:
            allRun = checkExecutable(args.toolPath)
            saveConfigFile(allRun, args.toolPath, greedyFasPath)

    if args.check:
        if not os.path.exists(os.path.abspath(greedyFasPath+'/pathconfig.txt')):
            sys.exit('ERROR: %s not found' % (greedyFasPath+'/pathconfig.txt'))
        else:
            with open(greedyFasPath+'/pathconfig.txt', 'r') as file:
                savedPath = file.read().strip()
                checkAnnoToolsFile(savedPath)
                print('Annotation tools can be found at %s. FAS is ready to run!' % savedPath)
                print('You should test fas.doAnno with this command:')
                print('==> fas.doAnno -i test_annofas.fa -o testFas_output <==')
                sys.exit()

    if args.savePath:
        checkAnnoToolsFile(args.toolPath)
        with open(greedyFasPath+'/pathconfig.txt', 'w') as config:
            config.write(os.path.abspath(args.toolPath))
            config.close()
        print('Annotation tools can be found at %s. FAS is ready to run!' % args.toolPath)
        print('You should test fas.doAnno with this command:')
        print('==> fas.doAnno -i test_annofas.fa -o testFas_output <==')
        print('*** NOTE: For using FAS you need to source %s/fas.profile first!' % args.toolPath)
        sys.exit()

    if args.addLength:
        if not os.path.exists(os.path.abspath(args.toolPath+'/annoTools.txt')):
            sys.exit('ERROR: %s not found' % (args.toolPath+'/annoTools.txt'))
        else:
            if not os.path.exists(os.path.abspath(f'{args.toolPath}/Pfam/Pfam-hmms/Pfam-A.hmm.length')):
                print(f'Creating length file for PFAM...')
                write_phmm_length(args.toolPath, 'Pfam')
            if not os.path.exists(os.path.abspath(f'{args.toolPath}/SMART/SMART-hmms/SMART.hmm.length')):
                print(f'Creating length file for SMART...')
                write_phmm_length(args.toolPath, 'SMART')
            sys.exit()


    anno_path = install_annoTool(options)
    allRun = checkExecutable(anno_path)
    saveConfigFile(allRun, anno_path, greedyFasPath)


if __name__ == '__main__':
    main()



#!/bin/env python

#######################################################################
# Copyright (C) 2019 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


from operator import itemgetter
import os
import multiprocessing
from functools import partial
from copy import deepcopy
from tqdm import tqdm
import sys
from time import sleep
from greedyFAS.mainFAS.greedyFAS import su_search_protein
from greedyFAS.mainFAS.fasInput import read_json, check_version
from greedyFAS.mainFAS.fasOutput import write_domain_out_fad, write_tsv_out, phyloprofile_out
from greedyFAS.mainFAS.fadScoring import fad_entire_calc_score
from greedyFAS.mainFAS.fasPathing import pb_region_mapper, pb_region_paths
from greedyFAS.annoFAS.annoModules import mergeNestedDic


# important vars #             ###  var looks ###
# naming
# seed_proteome = {}           #{("protein_id", {("domain_name", [("START", "STOP")])})}
# query_proteome = {}          #{("protein_id", {("domain_name", [("START", "STOP")])})}
# clan_dict = {}               #{("domain_name", "clan")}
# search_features = {}         #{("F_0", ("domain_name", "POSITION", "Start", "Stop" ))}
# a_s_f = {}                   # additional seed features [non linearized]
# query_features = {}          #{("F_0", ("domain_name", "POSITION", "Start", "Stop" ))}
# a_q_f = {}                   # additional query features [non linearized]
# query_protein = {}           #{("domain_name", ["POSITION_1", "POSITION_2"])}
# query_clans = {}             #{"clan": "INSTANCES"}
# weights = {}                 #{"domain_name": "weight"}


# flow control functions


def fc_start(option):
    """Overhead function,
    this function manages the individual functions that read the input files and prepares the data for the main script.

    :param option: dictionary that contains the main option variables of FAD
    """
    clan_dict = {}
    option["reverse"] = False
    interprokeys = {}
    phmm = {}
    v_warning = False
    version = None
    if option['weight_path']:
        weights = read_json(option['weight_path'])
    else:
        weights = {'weights': {}, 'default': 0.5}
    proteome_list = []
    for path in option["p_path"]:
        proteome = read_json(path)
        proteome_list.append(proteome["feature"])
        clan_dict.update(proteome["clan"])
        if 'interproID' in proteome:
            interprokeys.update(proteome['interproID'])
        if 'length' in proteome:
            phmm.update(proteome['length'])
        version, v_warning = check_version(version, proteome, v_warning)
    seed_proteome = mergeNestedDic(proteome_list)
    proteome_list = []
    for path in option["s_path"]:
        proteome = read_json(path)
        proteome_list.append(proteome["feature"])
        clan_dict.update(proteome["clan"])
        if 'interproID' in proteome:
            interprokeys.update(proteome['interproID'])
        if 'length' in proteome:
            phmm.update(proteome['length'])
        version, v_warning = check_version(version, proteome, v_warning)
    query_proteome = mergeNestedDic(proteome_list)
    if v_warning:
        print('##########\nWARNING: There are version discrepancies in the annotation files. This means that there '
              'might be differences in tool/database versions of the annotation. You should check the annotation files '
              'with "fas.getAnnoVersion"!\n##########')
    for tool in option["input_linearized"]:
        if tool not in seed_proteome[list(seed_proteome)[0]]:
            raise Exception(tool + " is missing in the seed annotation")
        if tool not in query_proteome[list(query_proteome)[0]]:
            raise Exception(tool + " is missing in the query annotation")
    if option["seed_id"]:
        for protid in option["seed_id"]:
            if protid not in seed_proteome:
                raise Exception(protid + " is not in the seed annotation")
    if option["query_id"]:
        for protid in option["query_id"]:
            if protid not in query_proteome:
                raise Exception(protid + " is not in the query annotation")
    if option["bidirectional"]:
        print("calculating forward scores...")
        f_results = fc_main(weights, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm)
        id_tmp = option["seed_id"]
        option["reverse"] = True
        option["seed_id"] = option["query_id"]
        option["query_id"] = id_tmp
        if option["pairwise"]:
            pairtmp = []
            for pair in option["pairwise"]:
                pairtmp.append((pair[1], pair[0]))
            option["pairwise"] = pairtmp
        print("calculating backward scores...")
        r_results = fc_main(weights, query_proteome, seed_proteome, clan_dict, option, interprokeys, phmm)
        if option["phyloprofile"]:
            phyloprofile_out(option["outpath"], True, option["phyloprofile"], (f_results, r_results))
        if not option['tsv']:
            write_tsv_out(option["outpath"], True, (f_results, r_results))
    else:
        print("calculating forward scores...")
        results = fc_main(weights, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm)
        if not option["tsv"]:
            write_tsv_out(option["outpath"], False, (results, None))
        if option["phyloprofile"]:
            phyloprofile_out(option["outpath"], False, option["phyloprofile"], [results, None])


def fc_main(weights, seed_proteome, query_proteome, clan_dict, option, interprokeys, phmm):
    """Main function,
    manages linearization and scoring, creates the output
    Function calls: su_lin_query_protein(), pb_graphtraversal(),
                    pb_entire_graphtraversal_priority(),
                    sf_entire_calc_score(), pb_entire_main_nongreedy(), w_weight_const_rescale()

    :param weighting: dictionary that contains the weights for each feature
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param query_proteome: dictionary that contains the feature architecture of all query proteins
    :param clan_dict: dictionary that maps features to clans
    :param option: dictionary that contains the main option variables of FAS
    """

    progress = None
    domain_out = None
    results = []
    if option["domain"]:
        if option["reverse"]:
            domain_out = open(option["outpath"] + "_reverse.domains", "w")
        else:
            domain_out = open(option["outpath"] + "_forward.domains", "w")
        domain_out.write('# pairID\torthoID\tseqLen\tfeature\tfStart\tfEnd\tfWeight\tfPath\tinterProID\te-value'
                         + '\tbitScore\tpStart\tpEnd\tpLen\n')
    if option['pairwise']:
        if option['progress']:
            progress = tqdm(total=len(option['pairwise']), file=sys.stdout)
        for pair in option['pairwise']:
            query = pair[1]
            protein = pair[0]
            if query not in query_proteome:
                raise Exception(query + ' is missing in annotation!')
            if protein not in seed_proteome:
                raise Exception(protein + ' is missing in annotation!')
            tmp_query = fc_prep_query(query, query_proteome, option, clan_dict)
            query_graph, all_query_paths, lin_query_set, query_features, a_q_f, query_clans, clan_dict, go_priority\
                = tmp_query[0:8]

            ####
            results.append(fc_main_sub(protein, weights, seed_proteome, option, all_query_paths, query_features,
                                       go_priority, a_q_f, clan_dict, query_graph, query_proteome, query, query_clans,
                                       domain_out, interprokeys, phmm))
            if option["progress"]:
                progress.update(1)
    else:
        if option["query_id"]:
            querylist = option["query_id"]
            for query in querylist:
                if query not in query_proteome:
                    raise Exception(query + ' is missing in annotation!')
        else:
            querylist = list(query_proteome.keys())
        if option["seed_id"]:
            seedlist = option["seed_id"]
            for seed in seedlist:
                if seed not in seed_proteome:
                    raise Exception(seed + ' is missing in annotation!')
        else:
            seedlist = list(seed_proteome.keys())
        if option['progress']:
            progress = tqdm(total=len(querylist)*len(seedlist), file=sys.stdout)
        for query in querylist:
            tmp_query = fc_prep_query(query, query_proteome, option, clan_dict)
            query_graph, all_query_paths, lin_query_set, query_features, a_q_f, query_clans, clan_dict, go_priority\
                = tmp_query[0:8]

            for protein in seedlist:
                results.append(fc_main_sub(protein, weights, seed_proteome, option, all_query_paths,
                                           query_features, go_priority, a_q_f, clan_dict, query_graph, query_proteome,
                                           query, query_clans, domain_out, interprokeys, phmm))
                if option["progress"]:
                    progress.update(1)
        if option["progress"]:
            progress.refresh()
            progress.close()
            sleep(1.0)
    return results


def fc_prep_query(query, query_proteome, option, clan_dict):
    go_priority = False
    lin_query_set, query_features, a_q_f, query_clans, clan_dict = su_lin_query_protein(
        query, query_proteome, clan_dict, option)
    tmp_query_graph, path_number = pb_region_paths(pb_region_mapper(
        lin_query_set, query_features, option["max_overlap"], option["max_overlap_percentage"]))
    # PRIORITY CHECK: checking for number of instances - assess complexity of the feature graph
    if (len(query_features) > option["priority_threshold"] or path_number > option["max_cardinality"]) and \
            option["priority_mode"]:
        go_priority = True
        all_query_paths = "PRIORITY"
    elif option["priority_mode"]:
        # creating all paths
        all_query_paths = pb_graphtraversal(tmp_query_graph, [], [], option)
    elif len(tmp_query_graph) == 0:
        all_query_paths = []
    else:
        all_query_paths = "NOPRIORITY"
    return tmp_query_graph, all_query_paths, lin_query_set, query_features, a_q_f, query_clans, clan_dict, \
        go_priority


def fc_main_sub(protein, weights, seed_proteome, option, all_query_paths, query_features, go_priority, a_q_f,
                clan_dict, tmp_query_graph, query_proteome, query, query_clans, domain_out, interprokeys, phmm):
    go_priority_2 = False
    mode = 0
    pathcount = 0
    unsolved = True
    search_protein, search_features, a_s_f = su_search_protein(protein, seed_proteome, option)
    search_protein = tuple(search_protein)
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    region = pb_region_mapper(list(search_protein), search_features, option["max_overlap"],
                              option["max_overlap_percentage"])
    search_graph, search_path_number = pb_region_paths(region)
    # check for available paths
    # query <--VS-- seed
    # case M2.1: empty(query)

    if int(len(all_query_paths)) == 0 or int(len(query_features) == 0):
        # case M2.1.1: empty(query)-empty(search)
        # should be the best fix independent from weight
        if int(len(search_features)) == 0:
            path = list(a_s_f.keys())
            query_architecture = list(a_q_f.keys())
            score_w = fad_entire_calc_score(path, query_architecture, weights, search_features, a_s_f,
                                           query_features, a_q_f, clan_dict, option)
            mode = 0
        else:
            # case M2.1.2: empty(query)-graph(search)
            tmp_path_score = pb_entire_main_nongreedy(protein, [], search_features, weights, query_features,
                                                      seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                                      option, search_graph, search_path_number)
            path = tmp_path_score[0][0]
            score_w = tmp_path_score[0][1]
            query_architecture = tmp_path_score[0][2]
            mode = 0

        # set max_fixture according to
        max_fixture = (path, score_w, query_architecture, protein)
        unsolved = False

    # seed architecture == query architecture
    elif seed_proteome[protein] == query_proteome[query]:
        path = list(a_s_f.keys()) + list(search_features.keys())
        query_architecture = list(a_q_f.keys()) + list(query_features.keys())
        score_w = fad_entire_calc_score(path, query_architecture, weights, search_features, a_s_f,
                                       query_features, a_q_f, clan_dict, option)
        mode = 0
        max_fixture = (path, score_w, query_architecture, protein)
        unsolved = False

    # handle all paths
    # case M2.2: graph(query)
    elif not option["priority_mode"] and not int(len(query_features)) == 0:
        pool_results = None
        mode = 0
        stack, jobpaths = pb_create_jobs(tmp_query_graph, option)
        if len(stack) > 0:
            jobpool = multiprocessing.Pool(processes=option["cores"])
            func = partial(pb_graph_traversal_sub, protein, search_features, weights, query_features, seed_proteome,
                           a_s_f, a_q_f, clan_dict, query_clans, tmp_query_graph, option, search_graph,
                           search_path_number)
            try:
                pool_results = jobpool.map_async(func, stack).get(timeout=option["timelimit"])
            except multiprocessing.TimeoutError:
                go_priority_2 = True
                jobpool.terminate()
            if not go_priority_2:
                for pool_result in pool_results:
                    if pool_result[1][3] >= max_fixture[1][3]:
                        max_fixture = deepcopy(pool_result)
        if len(jobpaths) > 0 and not go_priority_2:
            # case M2.2.1: graph(query)-empty(search)
            if int(len(search_features)) == 0:
                for jobpath in jobpaths:
                    pathcount += 1
                    # special case: protein with no pfam or smart domains
                    # get score for a_s_f and query_path directly
                    path = list(a_s_f.keys())
                    query_path_ad = jobpath + list(a_q_f.keys())
                    score_w = fad_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)

                    # check for max scoring fixture of path and query_path
                    if (score_w[4] >= max_fixture[1][4] and score_w[3] == max_fixture[1][3]) or \
                            score_w[3] >= max_fixture[1][3]:
                        max_fixture = (path, score_w, query_path_ad, protein)
            else:
                # case M2.2.2 graph(query)-graph(search)
                # regular traversal of graph based on search_protein
                jobpool = multiprocessing.Pool(processes=option["cores"])
                jobs = []
                jobcount = (int(len(jobpaths) / option["cores"]), len(jobpaths) % option["cores"])
                if jobcount[0] == 0:
                    for i in range(jobcount[1]):
                        jobs.append([jobpaths[-(i + 1)]])
                else:
                    counter = 0
                    for i in range(option["cores"]):
                        jobs.append(jobpaths[counter: (i + 1) * jobcount[0]])
                        counter = (i + 1) * jobcount[0]
                    for i in range(jobcount[1]):
                        jobs[i].append(jobpaths[-(i + 1)])
                func = partial(pb_calc_sub, protein, search_features, weights, query_features,
                               seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option, search_graph,
                               search_path_number)
                try:
                    pool_results = jobpool.map_async(func, jobs).get(timeout=option["timelimit"])
                except multiprocessing.TimeoutError:
                    go_priority_2 = True
                    jobpool.terminate()
                if not go_priority_2:
                    for pool_result in pool_results:
                        if pool_result[1][3] >= max_fixture[1][3]:
                            max_fixture = deepcopy(pool_result)
    if go_priority or go_priority_2 and unsolved:
        mode = 1
        all_query_paths = []
        priority_list = []
        for tool in option["input_linearized"]:
            for domain_type in query_proteome[query][tool]:
                priority_list.append(domain_type)
        priority_list.append("NONE")
        for domain_type in priority_list:
            all_query_paths += (
                pb_entire_graphtraversal_priority(tmp_query_graph, domain_type, protein, 1, search_features,
                                                  weights, query_features, seed_proteome, a_s_f, a_q_f,
                                                  clan_dict, query_clans, option))
            # max fixture of (search_path, score, query_path)
    if option["priority_mode"] or go_priority_2 and unsolved:
        for query_path in all_query_paths:
            pathcount += 1

            # case M2.2.1: graph(query)-empty(search)
            if int(len(search_features)) == 0:
                # special case: protein with no pfam or smart domains
                # get score for a_s_f and query_path directly
                path = list(a_s_f.keys())
                query_path_ad = query_path + list(a_q_f.keys())
                score_w = fad_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                               query_features, a_q_f, clan_dict, option)
            else:
                # case M2.2.2 graph(query)-graph(search)
                # regular traversal of graph based on search_protein
                tmp_path_score = pb_entire_main_nongreedy(protein, query_path, search_features, weights,
                                                              query_features, seed_proteome, a_s_f, a_q_f, clan_dict,
                                                              query_clans, option, search_graph, search_path_number)

                path = tmp_path_score[0][0]
                score_w = tmp_path_score[0][1]
                query_path_ad = tmp_path_score[0][2]
                mode = tmp_path_score[1]

            # check for max scoring fixture of path and query_path
            if (score_w[4] >= max_fixture[1][4] and score_w[3] == max_fixture[1][3]) or \
                    score_w[3] >= max_fixture[1][3]:
                max_fixture = (path, score_w, query_path_ad, protein)

    score = max_fixture[1]
    try:
        if mode == 1:
            mode_out = "priority"
        else:
            mode_out = "exhaustive"
    except KeyError:
        mode_out = "default"
    if go_priority or go_priority_2:
        mode_out += "/priority"
    else:
        mode_out += "/exhaustive"

    best_template_path = []
    best_query_path = []

    for feature in max_fixture[0]:
        if feature in search_features:
            best_template_path.append((search_features[feature][0], search_features[feature][1],
                                       search_features[feature][2], search_features[feature][3]))
        else:
            best_template_path.append(
                (a_s_f[feature][0], a_s_f[feature][1], a_s_f[feature][2], a_s_f[feature][3]))

    for feature in max_fixture[2]:
        if feature in query_features:
            best_query_path.append((query_features[feature][0], query_features[feature][1],
                                    query_features[feature][2], query_features[feature][3]))
        else:
            best_query_path.append((a_q_f[feature][0], a_q_f[feature][1], a_q_f[feature][2], a_q_f[feature][3]))
    path_tmp = {}
    path_tmp_query = {}
    scale = max_fixture[1][7]
    for feature in best_template_path:
        if feature[0] in path_tmp:
            path_tmp[feature[0]].append((feature[2], feature[3]))
        else:
            path_tmp[feature[0]] = [(feature[2], feature[3])]
    for feature in best_query_path:
        if feature[0] in path_tmp_query:
            path_tmp_query[feature[0]].append((feature[2], feature[3]))
        else:
            path_tmp_query[feature[0]] = [(feature[2], feature[3])]

    # unweighted case
    if option["domain"]:
        write_domain_out_fad(seed_proteome, query_proteome, protein, query, weights, scale, path_tmp, path_tmp_query,
                             domain_out, option, interprokeys, phmm)
    if option["raw"]:
        print('#' + '\t' + protein + '\t' + query + '\t' + str(score[3]))
    return protein, query, (score[3], score[0], score[1], score[2], score[6]), mode_out


# Start Up Functions <su>


def su_lin_query_protein(protein_id, query_proteome, clan_dict, option):
    """Initializes variables for the current query protein (linearization version)

    :param option: dictionary that contains the main option variables of FAS
    :param protein_id: String that contains the identifier of the query protein
    :param query_proteome: dictionary that contains the feature architecture of all query proteins
    :param clan_dict: dictionary that maps features to clans
    :return:lin_query_protein, query_features, a_q_f(additional[not linearized] query features), query_clans, clan_dict
    """

    lin_query_protein = []
    query_clans = {}
    query_features = {}
    a_q_f = {}
    tmp = []
    i = 0

    for tool in option["input_linearized"]:
        for feature in query_proteome[protein_id][tool]:
            e_feature = False
            try:
                if query_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                clan = None
                if feature in clan_dict:
                    clan = clan_dict[feature]
                tmp_clan = 0
                for instance in query_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            query_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "F_" + str(i)
                        query_features[key] = (feature, position, instance[0], instance[1])
                        tmp.append((key, instance[1]))
                        i += 1
                        tmp_clan += 1
                if clan in query_clans:
                    query_clans[clan] += tmp_clan
                elif tmp_clan > 0:
                    query_clans[clan] = tmp_clan
    for tool in option["input_normal"]:
        for feature in query_proteome[protein_id][tool]:
            e_feature = False
            try:
                if query_proteome[protein_id][tool][feature]["evalue"] <= option["eFeature"]:
                    e_feature = True
            except TypeError:
                e_feature = True
            if e_feature:
                clan = None
                if feature in clan_dict:
                    clan = clan_dict[feature]
                tmp_clan = 0
                for instance in query_proteome[protein_id][tool][feature]["instance"]:
                    e_instance = False
                    try:
                        if instance[2] <= option["eInstance"]:
                            e_instance = True
                    except TypeError:
                        e_instance = True
                    if e_instance:
                        position = ((float(instance[0]) + float(instance[1])) / 2.0) / float(
                            query_proteome[protein_id]["length"])
                        position = round(position, 4)
                        key = "O_" + str(i)
                        a_q_f[key] = (feature, position, instance[0], instance[1])
                        i += 1
                        tmp_clan += 1
                if clan in query_clans:
                    query_clans[clan] += tmp_clan
                elif tmp_clan > 0:
                    query_clans[clan] = tmp_clan

    # sort  instances
    tmp2 = sorted(tmp, key=itemgetter(1))
    for x in tmp2:
        lin_query_protein.append(x[0])

    return lin_query_protein, query_features, a_q_f, query_clans, clan_dict


# Path-building Functions # <pb>
# Used for the Pfam/Smart domains (default)
def pb_create_jobs(graph, option):
    """Does a breadth-first search to divide the feature architecture graph into several sub-problems so that it can be
    worked on by multiple cores. There are more jobs created than there are cores to counter the possibility of a core
    having nothing to do because of different running times for each job.

    :param graph: graph containing all paths through a protein architecture
    :param option: dictionary that contains the main option variables of FAS
    :return: queue (contains the starting points and sub-paths of the jobs), paths (empty unless b-f search reaches end
             node)
    """
    limit = 4 * option["cores"]
    queue = [("START", [])]
    paths = []
    while len(queue) < limit and queue:
        vertex, path = queue.pop(0)
        for next_vertex in graph[vertex]:
            if next_vertex == "END":
                paths.append(path)
            else:
                queue.append((next_vertex, path + [next_vertex]))
    return queue, paths


def pb_graph_traversal_sub(protein, search_features, weights, query_features, seed_proteome, a_s_f, a_q_f, clan_dict,
                           query_clans, query_graph, option, search_graph, search_path_number, stack):
    """Sub function for multi-processing: This function does a depth-first search in the feature graph. The starting
     point in the graph is given in the stack.

    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :param protein: String that contains the identifier of the seed protein
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param stack: contains the starting point in the graph and sub-path
    :param query_graph: graph of the query architecture
    :return: max_fixture
    """
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    v_stack, p_stack = [stack[0]], [stack[1]]
    while v_stack:
        query_path, v_stack, p_stack = pb_graphtraversal(query_graph, v_stack, p_stack, option)
        # case M2.2.1: graph(query)-empty(search)
        if int(len(search_features)) == 0:
            # special case: protein with no pfam or smart domains
            # get score for a_s_f and query_path directly
            path = list(a_s_f.keys())
            query_path_ad = query_path + list(a_q_f.keys())
            score_w = fad_entire_calc_score(path, query_path_ad, weights, search_features, a_s_f,
                                            query_features, a_q_f, clan_dict, option)
        else:
            # case M2.2.2 graph(query)-graph(search)
            # regular traversal of graph based on search_protein
            tmp_path_score = pb_entire_main_nongreedy(protein, query_path, search_features, weights, query_features,
                                                      seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option,
                                                      search_graph, search_path_number)
            path = tmp_path_score[0][0]
            score_w = tmp_path_score[0][1]
            query_path_ad = tmp_path_score[0][2]

        # check for max scoring fixture of path and query_path
        if ((not max_fixture[1][5]) and (score_w[4] >= max_fixture[1][4]) and score_w[3] <= max_fixture[1][3]) or \
                score_w[3] >= max_fixture[1][3]:
            max_fixture = (path, score_w, query_path_ad, protein)
    return max_fixture


def pb_calc_sub(protein, search_features, weights, query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                option, search_graph, search_path_number, jobpaths):
    """Sub function for multiprocessing [2]: Goes through a list of (query) paths an evaluates them.

    :param protein: String that contains the identifier of the seed protein
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param jobpaths: a group of (query) paths to be evaluated
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :return: max_fixture, tmpcheck - tmpstart (time taken for calculation)

    """
    max_fixture = ([], (0.0, 0.0, 0.0, 0.0, 0, 0, False), [], protein)
    for jobpath in jobpaths:
        tmp_path_score = pb_entire_main_nongreedy(protein, jobpath, search_features, weights, query_features,
                                                  seed_proteome, a_s_f, a_q_f, clan_dict, query_clans, option,
                                                  search_graph, search_path_number)
        path = tmp_path_score[0][0]
        score_w = tmp_path_score[0][1]
        query_path_ad = tmp_path_score[0][2]
        if ((not max_fixture[1][5]) and (score_w[4] >= max_fixture[1][4]) and score_w[3] <= max_fixture[1][3]) or \
                score_w[3] >= max_fixture[1][3]:
            max_fixture = (path, score_w, query_path_ad, protein)
    return max_fixture


def pb_entire_main_nongreedy(protein_id, query_path, search_features, weights, query_features, seed_proteome, a_s_f,
                             a_q_f, clan_dict, query_clans, option, search_graph, search_path_number):
    """Main Path-building function,
    creates graph with all paths(for seed/search protein), checks priority mode activation if necessary retrieves best
    path from graph traversal function, returns best path, score and mode
    Function calls: pb_region_mapper(), pb_region_paths(), pb_entire_priority_mode(),
                    pb_entire_graphtraversal()


    :param protein_id: String that contains the identifier of the seed protein
    :param query_path: currently evaluated query path
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :param search_path_number: number of paths in search protein
    :param search_graph: architecture graph of the search protein
    :return: path_score, mode[priority or extensive]

    """
    if (int(len(search_features)) >= option["priority_threshold"] or search_path_number > option['max_cardinality']) \
            and option["priority_mode"]:
        mode = 1
        path_score = pb_entire_priority_mode(protein_id, query_path, search_graph, search_features, weights,
                                             query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                             option)
    else:
        mode = 0
        path_score = pb_entire_graphtraversal(search_graph, query_path, search_features, weights, query_features,
                                              a_s_f, a_q_f, clan_dict, option)
    return path_score, mode


def pb_entire_priority_mode(protein, query_path, search_graph, search_features, weights, query_features, seed_proteome,
                            a_s_f, a_q_f, clan_dict, query_clans, option):
    """Path-evaluation in priority mode

    :param protein: String that contains the identifier of the seed protein
    :param query_path: currently evaluated query path
    :param search_graph: graph containing all paths through the seed architecture
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :return: best_path
    """

    # best_path (Path, (MS_score(float), PS_score(float), CS_score(float), final_score(float), path_weight(float),
    # common_feature(bool)), QPath)
    best_path = ("NULL", (0.0, 0.0, 0.0, 0.0, 0.0, False), "NULL")
    priority_list = []
    for tool in option["input_linearized"]:
        for domain_type in seed_proteome[protein][tool]:
            priority_list.append(domain_type)
    priority_list.append("NONE")
    for domain_type in priority_list:
        path_score_w = pb_entire_graphtraversal_priority(search_graph, domain_type, query_path, 0, search_features,
                                                         weights, query_features, seed_proteome, a_s_f, a_q_f,
                                                         clan_dict, query_clans, option)
        if path_score_w[1][5]:
            if path_score_w[1][3] >= best_path[1][3]:
                best_path = (path_score_w[0], path_score_w[1], path_score_w[2])
        else:
            # if so far no path with common features found AND the weight is higher
            if (not best_path[1][5]) and (path_score_w[1][4] >= best_path[1][4]):
                best_path = (path_score_w[0], path_score_w[1], path_score_w[2])
    return best_path


def pb_entire_graphtraversal(search_graph, query_path, search_features, weights, query_features, a_s_f, a_q_f,
                             clan_dict, option):
    """Traverses the feature architecture graph (seed) to score all paths

    :param search_graph: graph containing all paths through the seed architecture
    :param query_path: currently evaluated query path
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param option: dictionary that contains the main option variables of FAS
    :return: best_path
    """

    v_stack = ["START"]
    p_stack = []
    best_path = ([], (0.0, 0.0, 0.0, 0.0, 0.0, False, 0.0), [])
    first = 1
    if len(query_path) > 0:
        while v_stack:
            vertex = v_stack.pop()
            if len(p_stack) == 0:
                path = []
            else:
                path = p_stack.pop()
            for next_vertex in search_graph[vertex]:
                if next_vertex == "END":
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = fad_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3]) or \
                       score_w[3] >= best_path[1][3]:
                        best_path = (path_ad, score_w, query_path_ad)
                else:
                    v_stack.append(next_vertex)
                    p_stack.append(path + [next_vertex])
    else:
        while v_stack and first:
            vertex = v_stack.pop()
            if len(p_stack) == 0:
                path = []
            else:
                path = p_stack.pop()
            for next_vertex in search_graph[vertex]:
                if next_vertex == "END":
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = fad_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                   query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3]) or \
                       score_w[3] >= best_path[1][3]:
                        best_path = (path_ad, score_w, query_path_ad)
                    first = 0
                else:
                    v_stack.append(next_vertex)
                    p_stack.append(path + [next_vertex])
    return best_path


def pb_graphtraversal(graph, v_stack, p_stack, option):
    """Traverses the feature architecture graph (query) to evaluate number of paths, returns all paths if # of paths is
     smaller than the max_cardinality

    :param graph: graph containing all paths through the (query) architecture
    :param option: dictionary that contains the main option variables of FAS
    :param v_stack: vertex stack
    :param p_stack: path stack
    :return: paths
    """
    paths = []
    if v_stack == [1]:
        v_stack = ["START"]
        p_stack = []
        mode = 1
    elif not v_stack:
        v_stack = ["START"]
        p_stack = []
        mode = 0
    else:
        mode = 1
    while v_stack:
        vertex = v_stack.pop()
        if len(p_stack) == 0:
            path = []
        else:
            path = p_stack.pop()
        for next_vertex in graph[vertex]:
            if next_vertex == "END":
                if mode == 0:
                    paths.append(path)
                    if len(paths) > option["max_cardinality"]:
                        return paths
                else:
                    return path, v_stack, p_stack
            else:
                v_stack.append(next_vertex)
                p_stack.append(path + [next_vertex])
    return paths


def pb_entire_graphtraversal_priority(search_graph, priority, query_path, mode, search_features, weights,
                                      query_features, seed_proteome, a_s_f, a_q_f, clan_dict, query_clans,
                                      option):
    """Traverses the feature architecture graph in priority mode

    :param search_graph: graph containing all paths through the seed architecture
    :param priority: current priority feature
    :param query_path: graph containing all paths through the query architecture
    :param mode: query or seed linearization
    :param search_features: all features to be linearized in the seed protein
    :param weights: feature weights
    :param query_features: all features to be linearized in the query protein
    :param seed_proteome: dictionary that contains the feature architecture of all seed proteins
    :param a_s_f: additional seed features [non linearized]
    :param a_q_f: additional query features [non linearized]
    :param clan_dict: dictionary that maps features to clans
    :param query_clans: (Pfam)-clans of the current query protein
    :param option: dictionary that contains the main option variables of FAS
    :return: paths or best_path
    """

    v_stack = ["START"]
    p_stack = []
    protein = None
    paths = []
    best_path = None
    if mode == 0:
        best_path = ([], (0.0, 0.0, 0.0, 0.0, 0.0, False), [])
    else:
        protein = query_path
    while v_stack:
        vertex = v_stack.pop()
        if len(p_stack) == 0:
            path = []
        else:
            path = p_stack.pop()

        p_found = 0
        p_candidates = []
        for next_vertex in search_graph[vertex]:
            if next_vertex == "END":
                if mode == 1:
                    paths.append(path)
                    p_found = 2
                else:
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_q_f.keys())
                    score_w = fad_entire_calc_score(path_ad, query_path_ad, weights, search_features, a_s_f,
                                                    query_features, a_q_f, clan_dict, option)
                    if ((score_w[4] >= best_path[1][4] and score_w[3] == best_path[1][3])
                            or score_w[3] >= best_path[1][3]):
                        best_path = (path_ad, score_w, query_path_ad)
                    p_found = 2
            elif mode == 1:
                if query_features[next_vertex][0] == priority:
                    p_candidates.append(next_vertex)
                    p_found = 1
            elif mode == 0:
                p_candidates.append(next_vertex)
                p_found = 1
        if p_found == 1:
            if len(p_candidates) == 1:
                v_stack.append(p_candidates[0])
                p_stack.append(path + p_candidates)
            else:
                if mode == 1:
                    best_priority_bridger = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                    for next_vertex in p_candidates:
                        ############ changed
                        score = fad_entire_calc_score(list(search_features.keys()) + list(a_s_f.keys()),
                                                      path + [next_vertex], weights, search_features, a_s_f,
                                                      query_features, a_q_f, clan_dict, option)
                        if score[3] >= best_priority_bridger[1][3]:
                            best_priority_bridger = (next_vertex, score[0:4])
                    v_stack.append(best_priority_bridger[0])
                    p_stack.append(path + [best_priority_bridger[0]])
                elif mode == 0:
                    # greedy strategy: if feature type priority appears more than once
                    best_partial_path = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                    for next_vertex in p_candidates:
                        path_ad = path + list(a_s_f.keys())
                        query_path_ad = query_path + list(a_q_f.keys())
                        score_w = fad_entire_calc_score(path_ad + [next_vertex], query_path_ad, weights,
                                                        search_features, a_s_f, query_features, a_q_f, clan_dict, option)
                        if (score_w[4] >= best_partial_path[1][4] and score_w[3] == best_partial_path[1][3]) \
                           or score_w[3] >= best_partial_path[1][3]:
                            best_partial_path = (next_vertex, score_w)

                    v_stack.append(best_partial_path[0])
                    p_stack.append(path + [best_partial_path[0]])

        elif p_found == 0:
            if mode == 1:
                best_priority_bridger = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                for next_vertex in search_graph[vertex]:
                    ############## changed
                    score = fad_entire_calc_score(list(search_features.keys()) + list(a_s_f.keys()),
                                                  path + [next_vertex], weights, search_features, a_s_f,
                                                  query_features, a_q_f, clan_dict, option)
                    if score[3] >= best_priority_bridger[1][3]:
                        best_priority_bridger = (next_vertex, score)
                v_stack.append(best_priority_bridger[0])
                p_stack.append(path + [best_priority_bridger[0]])

            elif mode == 0:
                # greedy strategy: if feature type (p priority) not found
                best_partial_path = ("NONE", (0.0, 0.0, 0.0, 0.0, 0.0, False))
                for next_vertex in search_graph[vertex]:
                    path_ad = path + list(a_s_f.keys())
                    query_path_ad = query_path + list(a_s_f.keys())
                    score_w = fad_entire_calc_score(path_ad + [next_vertex], query_path_ad, weights, search_features,
                                                   a_s_f, query_features, a_q_f, clan_dict, option)
                    if (score_w[4] >= best_partial_path[1][4] and score_w[3] == best_partial_path[1][3]) or \
                       score_w[3] >= best_partial_path[1][3]:
                        best_partial_path = (next_vertex, score_w)

                v_stack.append(best_partial_path[0])
                p_stack.append(path + [best_partial_path[0]])
    if mode == 1:
        return paths
    elif mode == 0:
        return best_path



#!/bin/env python

#######################################################################
# Copyright (C) 2019 Julian Dosch
#
# This file is part of greedyFAS.
#
#  greedyFAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  greedyFAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with greedyFAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

from setuptools import setup, find_packages
import re

with open('README.md', 'r') as input:
    long_description = input.read()

setup(
    name='greedyFAS',
    version='1.17.13',
    python_requires='>=3.7.0',
    description='A tool to compare protein feature architectures',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Julian Dosch',
    author_email='Dosch@bio.uni-frankfurt.de',
    url='https://github.com/BIONF/FAS',
    packages=find_packages(),
    package_data={'': ['*']},
    install_requires=[
        'biopython',
        'tqdm',
        'graphviz',
        'gnureadline',
        'GitPython',
        'pathlib'
    ],
    entry_points={
        'console_scripts': ['fas.parseAnno = greedyFAS.annoFAS.annoParserFAS:main',
                            'fas.doAnno = greedyFAS.annoFAS.annoFAS:main',
                            'fas.checkAnno = greedyFAS.annoFAS.checkAnno:main',
                            'fas.updateAnnoFile = greedyFAS.annoFAS.updateAnnoFile:main',
                            'fas.getProtByAnno = greedyFAS.annoFAS.getProtByAnno:main',
                            'fas.setup = greedyFAS.setupFAS:main',
                            'fas.run = greedyFAS.calcFAS:main',
                            'fas.runFdogFas = greedyFAS.fdogFAS:main',
                            'fas.calcComplexity = greedyFAS.complexityFAS:main',
                            'fas.getDomains = greedyFAS.domainFAS:main',
                            'fas.splitJson = greedyFAS.extractAnnoFAS:main',
                            'fas.disorder.predict = greedyFAS.disorderFAS.predict_disorder:main',
                            'fas.disorder.setup = greedyFAS.disorderFAS.install_aucpred:main',
                            'fas.mergeAnno = greedyFAS.mergeAnno:main',
                            'fas.getAnnoVersion = greedyFAS.annoFAS.getAnnoVersion:main',
                            'fas.overlapStatistics = greedyFAS.getOverlapStats:main',
                            'fad.run = greedyFAS.calcFAD:main'
                            ],
    },
    license='GPL-3.0',
    classifiers=[
        'Environment :: Console',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
)



#!/bin/env python

#######################################################################
# Copyright (C) 2019 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import json


def featuretypes(path):
    """
    Input function,
    reads the tools/featuretypes input-file and stores information in option

    :param path: path to input-file
    :param option: dictionary that contains the main option variables of FAS
    :return: option
    """
    input_linearized = []
    input_normal = []
    ifile = open(path, "r")
    lines = ifile.readlines()
    mode = "NULL"
    for line in lines:
        tmp = line.rstrip("\n").lower()
        if tmp == "#linearized":
            mode = "lin"
        elif tmp == "#normal":
            mode = "nor"
        elif tmp == "#checked":
            mode = "ignore"
        elif mode == "NULL":
            raise Exception(path + " is not a valid input file")
        elif mode == "lin" and len(tmp) > 0:
            input_linearized.append(tmp)
        elif mode == "nor" and len(tmp) > 0:
            input_normal.append(tmp)
    ifile.close()
    return input_linearized, input_normal


def constraints_in(path):
    """
    Input function,
    reads the constraints file

    :param path: path to input-file
    :return: constraints
    """
    constraints = {}
    cfile = open(path, "r")
    lines = cfile.readlines()
    i = 1
    if lines[0][0] == "#":
        while lines[i][0] != "#":
            split = (lines[i].rstrip("\n")).split("\t")
            if split[1] != "N":
                constraints[split[0]] = float(split[1])
            i += 1
    else:
        raise Exception(path + " might be in the wrong format. Please see the sample file in config directory.")
    if lines[i][0] == "#":
        i += 1
        while i < len(lines):
            split = (lines[i].rstrip("\n")).split("\t")
            constraints[split[0]] = float(split[1])
            i += 1
    cfile.close()
    return constraints


def check_version(version, proteome, v_warning):
    if 'version' in proteome:
        if version and not proteome['version'] == version:
            v_warning = True
        elif not version:
            version = proteome['version']
    else:
        if version and not version == 'NA':
            v_warning = True
        elif not version:
            version = 'NA'
    return version, v_warning


def read_pairwise(path):
    with open(path, 'r') as infile:
        pairwise = []
        line = infile.readline()
        while line:
            pairwise.append(line.rstrip('\n').split('\t'))
            line = infile.readline()
    return pairwise


def read_json(path):
    with open(path, 'r') as infile:
        in_dict = json.loads(infile.read())
    return in_dict



#!/bin/env python

#######################################################################
# Copyright (C) 2020 Julian Dosch
#
# This file is part of FAS.
#
#  FAS is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  FAS is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with FAS.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################


import os
import argparse
import multiprocessing as mp
from sys import argv
from greedyFAS.annoFAS import annoFAS
from greedyFAS.annoFAS import annoModules
from greedyFAS.mainFAS import fasInput, fasOutput, fadMain
from greedyFAS.calcFAS import anno
from pkg_resources import get_distribution


def get_options():
    version = get_distribution('greedyFAS').version
    parser = argparse.ArgumentParser(description='You are running FAS version ' + str(version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/FAS/wiki")
    required = parser.add_argument_group('required arguments')
    general = parser.add_argument_group('general arguments')
    inargs = parser.add_argument_group('input arguments')
    outargs = parser.add_argument_group('output arguments')
    annotation = parser.add_argument_group('annotation arguments')
    weighting = parser.add_argument_group('weighting arguments')
    thresholds = parser.add_argument_group('threshold arguments')
    obscure = parser.add_argument_group('obscure arguments')
    parser.add_argument('--version', action='version', version=str(version))
    required.add_argument("-s", "--seed", default=None, type=str, required=True,
                          help="path to seed protein fasta file")
    required.add_argument("-q", "--query", default=None, type=str, required=True,
                          help="path to query protein fasta file")
    required.add_argument("-a", "--annotation_dir", default=None, type=str, required=True,
                          help='working directory, all annotations are be stored here')
    required.add_argument("-o", "--out_dir", default=None, type=str, required=True,
                          help="output directory, all outputfiles will be stored here")
    general.add_argument("--bidirectional", action="store_true",
                         help="calculate both scoring directions")
    general.add_argument('--cpus', help='number of cores', type=int, action='store', default=0)
    annotation.add_argument('--force', help='Force override annotations', action='store_true')
    annotation.add_argument("-f", "--eFeature", default="0.001", type=float,
                            help="eValue cutoff for PFAM/SMART domain, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument("-i", "--eInstance", default="0.01", type=float,
                            help="eValue cutoff for PFAM/SMART instances, applied during annotation but also during "
                                 "calculation")
    annotation.add_argument('--eFlps', help='eValue cutoff for fLPS', action='store', default=0.0000001, type=float)
    annotation.add_argument('--org', help='Organism of input sequence(s) for SignalP search',
                            choices=['euk', 'gram+', 'gram-'], action='store', default='euk', type=str)
    annotation.add_argument("--toolPath", dest="toolPath", default='', type=str,
                            help="Path to Annotion tool directory created with fas.setup")
    weighting.add_argument("-r", "--weights", default=None, type=str,
                           help="Path to file containing weighting. [json] format")
    inargs.add_argument("--query_id", default=None, nargs='*', type=str,
                        help="Choose specific proteins (ids divided by spaces) from the query input for calculation, "
                             "by default this is off (all proteins are used for calculation)")
    inargs.add_argument("--seed_id", default=None, nargs='*', type=str,
                        help="Choose specific proteins (ids divided by spaces) from the seed input for calculation, "
                             "by default this is off (all proteins are used for calculation)")
    inargs.add_argument("--pairwise", dest="pairwise", default=None, type=str,
                        help="deactivate all against all comparison, needs a pairing file with the ids that should be"
                             " compared (one pair per line tab seperated), please look at the FAS wiki pages for "
                             "templates")
    inargs.add_argument("-d", "--featuretypes", default=None, type=str,
                        help="inputfile that contains the tools/databases used to predict features. Please look at the "
                             "FAS wiki pages for templates of the the featuretypes input file")
    inargs.add_argument("--extra_annotation", default=None, nargs='*', type=str,
                        help="give naming conventions for extra annotation files, eg. if the file name for the seed is "
                             "someseed.json than the extra annotation files should be named someseed_"
                             "[EXTRA_ANNOTATION].json, these extra files need to exist for seed, query and the "
                             "references (if given)")
    outargs.add_argument("-n", "--out_name", default=None, type=str,
                         help="name for outputfiles, if none is given the name will be created from the seed and "
                              "query names")
    outargs.add_argument("--raw", dest="raw", action="store_true",
                         help="print FAS scores to terminal")
    outargs.add_argument("--tsv", dest="tsv", action="store_true",
                         help="deactivates creation of the tsv output, either use together with --raw or "
                              "--phyloprofile")
    outargs.add_argument("--phyloprofile", dest="phyloprofile", default=None, type=str,
                         help="activate phyloprofile output, needs mapping file for all query proteins")
    outargs.add_argument("--domain", dest="domain", action="store_false",
                         help="deactivate .domains output")
    outargs.add_argument("--no_config", dest="config", action="store_false",
                         help="deactivate *_config.yml output")
    thresholds.add_argument("-c", "--max_overlap", dest="max_overlap", default=0, type=int,
                            help="maximum size overlap allowed, default is 0 amino acids")
    thresholds.add_argument("--max_overlap_percentage", dest="max_overlap_percentage", default=0.4, type=float,
                            help="defines how much percent of a feature the overlap is allowed to cover, default "
                                 "is 0.4 (40%%)")
    thresholds.add_argument("-t", "--priority_threshold", type=int, default=30,
                            help="Change to define the feature number threshold for activating priority mode in the "
                                 "path evaluation. default=30")
    thresholds.add_argument("-m", "--max_cardinality", default=500, type=int,
                            help="Change to define the threshold for the maximal cardinality (number) of feature paths "
                                 "in a graph. If max. cardinality is exceeded the priority mode will be used to for "
                                 "the path evaluation. default=500")
    obscure.add_argument("--priority_mode", action='store_false',
                         help="deactivates the greedy strategy priority mode for larger architectures, NOT RECOMMENDED")
    obscure.add_argument("--timelimit", default=3600, type=int,
                         help="Sets a soft time limit in seconds for the calculation between a pair of proteins,"
                              "This limit does not necessarily represent the actual runtime. It only stops the "
                              "exhaustive strategy through the architecture and activate priority mode. "
                              "This option is only relevant if the option [--priority_mode] is set")
    obscure.add_argument("-w", "--score_weights", nargs=3, default=[0.7, 0.0, 0.3], type=float,
                         help="Defines how the three scores MS, CS and PS are weighted, takes three float arguments, "
                              "sum should be 1.0, the default is 0.7, 0.0, 0.3")
    args = parser.parse_args()
    return args


def fad(args, toolpath):
    option_dict = {"seed_id": args.seed_id, "query_id": args.query_id, "priority_mode": args.priority_mode,
                   "priority_threshold": args.priority_threshold, "max_cardinality": args.max_cardinality,
                   "cores": int(args.cpus), "raw": args.raw, "bidirectional": args.bidirectional,
                   "max_overlap": args.max_overlap, "timelimit": args.timelimit, "phyloprofile": args.phyloprofile,
                   "score_weights": [], "tsv": args.tsv, "max_overlap_percentage": 0.0, "domain": args.domain,
                   "pairwise": None, "eInstance": args.eInstance, "eFeature": args.eFeature, "progress": True,
                   "weight_path": args.weights}

    seedname = '.'.join(args.seed.split('/')[-1].split('.')[:-1])
    option_dict["p_path"] = [args.annotation_dir + '/' + seedname + '.json']
    queryname = '.'.join(args.query.split('/')[-1].split('.')[:-1])
    option_dict["s_path"] = [args.annotation_dir + '/' + queryname + '.json']
    if args.raw:
        option_dict["progress"] = False
    if option_dict["cores"] == 0:
        option_dict["cores"] = mp.cpu_count()-1
    if args.extra_annotation:
        for i in args.extra_annotation:
            option_dict["p_path"].append(args.annotation_dir + '/' + i + '.json')
            option_dict["s_path"].append(args.annotation_dir + '/' + i + '.json')
    try:
        test = 0.0
        for weight in args.score_weights:
            option_dict["score_weights"].append(weight)
            test += float(weight)
        if test != 1.0:
            raise Exception("The sum of all values in -w [--score_weights] should be 1.0")
        option_dict["score_weights"] = tuple(option_dict["score_weights"])
    except ValueError:
        raise Exception(str(args.score_weights) + " is not a valid input for -w [--score_weights]")
    if 0.0 <= float(args.max_overlap_percentage) <= 1.0:
        option_dict["max_overlap_percentage"] = float(args.max_overlap_percentage)
    else:
        raise Exception("[--max_overlap_percentage] should be between 0.0 and 1.0")
    if args.out_name:
        option_dict['outpath'] = args.out_dir.rstrip('/') + '/' + args.out_name
    else:
        option_dict['outpath'] = args.out_dir.rstrip('/') + '/' + seedname + '_' + queryname
    if args.featuretypes:
        option_dict['input_linearized'], option_dict['input_normal'] = fasInput.featuretypes(args.featuretypes)
    else:
        option_dict['input_linearized'], option_dict['input_normal'] = fasInput.featuretypes(toolpath
                                                                                             + '/' + 'annoTools.txt')
    if args.pairwise:
        option_dict["pairwise"] = fasInput.read_pairwise(args.pairwise)
    else:
        option_dict["pairwise"] = None
    print('Calculating FAD score...')
    fadMain.fc_start(option_dict)
    if args.config:
        fasOutput.write_metadata(option_dict['outpath'] + '_config.yml', args, ' '.join(argv),
                                 str(get_distribution('greedyFAS').version))
    print('done!')


def main():
    args = get_options()
    toolpath = args.toolPath
    if toolpath == '':
        pathconfigfile = os.path.realpath(__file__).replace('calcFAD.py', 'pathconfig.txt')
        with open(pathconfigfile) as f:
            toolpath = f.readline().strip()
    else:
        toolpath = os.path.abspath(args.toolPath)
    if not os.path.isdir(args.annotation_dir):
        os.mkdir(args.annotation_dir)
    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)
    annojobs = [args.seed]
    if args.query not in annojobs:
        annojobs.append(args.query)
    anno(annojobs, args, toolpath)
    fad(args, toolpath)


if __name__ == '__main__':
    main()
